# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"Dashboard": "core.dashboards.ipynb",
         "BaseGallery": "core.dashboards.ipynb",
         "RecordDastasetGallery": "core.dashboards.ipynb",
         "DatasetOverview": "core.dashboards.ipynb",
         "MultiDatasetOverview": "core.dashboards.ipynb",
         "DatasetComparison": "core.dashboards.ipynb",
         "DatasetFilter": "core.dashboards.ipynb",
         "DatasetFilterWithRangeSliderAndMultiSelect": "core.dashboards.ipynb",
         "DatasetFilterWithScatter": "core.dashboards.ipynb",
         "DatasetGenerator": "core.dashboards.ipynb",
         "DatasetGeneratorScatter": "core.dashboards.ipynb",
         "Observable": "core.data.ipynb",
         "ObservableList": "core.data.ipynb",
         "DatasetDescriptor": "core.data.ipynb",
         "StringDescriptor": "core.data.ipynb",
         "GenericDataset": "core.data.ipynb",
         "ObjectDetectionDatasetOverview": "dashboards.ipynb",
         "ObjectDetectionDatasetComparison": "dashboards.ipynb",
         "ObjectDetectionDatasetGeneratorScatter": "dashboards.ipynb",
         "ObjectDetectionDatasetGeneratorRange": "dashboards.ipynb",
         "ObjectDetectionResultOverview": "dashboards.ipynb",
         "InstanceSegmentationDatasetOverview": "dashboards.ipynb",
         "InstanceSegmentationDatasetComparison": "dashboards.ipynb",
         "InstanceSegmentationDatasetGeneratorScatter": "dashboards.ipynb",
         "InstanceSegmentationDatasetGeneratorRange": "dashboards.ipynb",
         "InstanceSegmentationResultOverview": "dashboards.ipynb",
         "RecordDataframeParser": "data.ipynb",
         "RecordDataset": "data.ipynb",
         "ResultsDataset": "data.ipynb",
         "BboxRecordDataframeParser": "data.ipynb",
         "DataDescriptorBbox": "data.ipynb",
         "StatsDescriptorBbox": "data.ipynb",
         "ImageStatsDescriptorBbox": "data.ipynb",
         "ClassStatsDescriptorBbox": "data.ipynb",
         "GalleryStatsDescriptorBbox": "data.ipynb",
         "BboxRecordDataset": "data.ipynb",
         "PrecisionRecallMetricsDescriptorObjectDetection": "data.ipynb",
         "ObjectDetectionResultsDataset": "data.ipynb",
         "InstanceSegmentationRecordDataframeParser": "data.ipynb",
         "DataDescriptorInstanceSegmentation": "data.ipynb",
         "StatsDescriptorInstanceSegmentation": "data.ipynb",
         "ImageStatsDescriptorInstanceSegmentation": "data.ipynb",
         "ClassStatsDescriptorInstanceSegmentation": "data.ipynb",
         "GalleryStatsDescriptorInstanceSegmentation": "data.ipynb",
         "InstanceSegmentationRecordDataset": "data.ipynb",
         "PrecisionRecallMetricsDescriptorInstanceSegmentation": "data.ipynb",
         "InstanceSegmentationResultsDataset": "data.ipynb",
         "APObjectDetection": "metrics.ipynb",
         "APInstanceSegmentation": "metrics.ipynb",
         "Filter": "plotting.controls.ipynb",
         "RangeFilter": "plotting.controls.ipynb",
         "CategoricalFilter": "plotting.controls.ipynb",
         "TimeFilter": "plotting.controls.ipynb",
         "ScatterFilter": "plotting.controls.ipynb",
         "GenericMulitScatterFilter": "plotting.controls.ipynb",
         "barplot": "plotting.core.ipynb",
         "histogram": "plotting.core.ipynb",
         "heatmap": "plotting.core.ipynb",
         "time_arc_plot": "plotting.core.ipynb",
         "table_from_dataframe": "plotting.core.ipynb",
         "stacked_hist": "plotting.core.ipynb",
         "categorical_2d_histogram": "plotting.core.ipynb",
         "categorical_2d_histogram_with_gui": "plotting.core.ipynb",
         "scatter_plot_with_gui": "plotting.core.ipynb",
         "plots_as_matrix": "plotting.core.ipynb",
         "toggle_legend_js": "plotting.utils.ipynb",
         "calculate_mixing_matrix": "plotting.utils.ipynb",
         "get_min_and_max_dates": "plotting.utils.ipynb",
         "convert_rgb_image_to_bokeh_rgb_image": "plotting.utils.ipynb",
         "draw_record_with_bokeh": "plotting.utils.ipynb",
         "erles_to_string": "utils.ipynb",
         "erles_to_counts_to_utf8": "utils.ipynb",
         "string_to_erles": "utils.ipynb",
         "correct_mask": "utils.ipynb",
         "decorrect_mask": "utils.ipynb"}

modules = ["core/dashboards.py",
           "core/data.py",
           "dashboards.py",
           "data.py",
           "metrics.py",
           "plotting/controls.py",
           "plotting/core.py",
           "plotting/utils.py",
           "utils.py"]

doc_url = "https://fstroth.github.io/icevision_dashboards/"

git_url = "https://github.com/fstroth/icevision_dashboards/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/core.data.ipynb (unless otherwise specified).

__all__ = ['Observable', 'ObservableList', 'DatasetDescriptor', 'StringDescriptor', 'GenericDataset']

# Cell
from typing import Union, Optional, Any, Iterable, Callable
import os
import shutil
from abc import ABC, abstractmethod

# Cell
class Observable(ABC):
    """Simple implementation of the observer pattern."""
    def __init__(self, callbacks=None):
        self._callbacks = [] if callbacks is None else callbacks

    def register_callback(self, callback: Callable):
        self._callbacks.append(callback)

    def trigger_callbacks(self):
        for callback in self._callbacks:
            callback(self)

# Cell
class ObservableList(Observable):
    """List with observer pattern. The internal list prepresentation can be accessed with the list attribute"""
    def __init__(self, observable_list: list, callbacks=None):
        self._list = observable_list
        super().__init__(callbacks)

    @property
    def list(self):
        return self._list

    @list.setter
    def list(self, value: Any):
        self._list = value
        self.trigger_callbacks()

    def __repr__(self):
        return self._list.__repr__()

    def __iter__(self):
        for item in self._list:
            yield item

    def __len__(self):
        return len(self._list)

    def __getitem__(self, index: int):
        return self._list[index]

    def __setitem__(self, index: int, value: Any):
        self._list[index] = value
        self.trigger_callbacks()

    def __add__(self, other):
        # turn the callbacks first into a set and than back into a list to avoid callbacks being triggered multiple times
        new_callbacks = list(set(self._callbacks+other._callbacks))
        return ObservableList(self.list + other.list, new_callbacks)

    def append(self, item: Any):
        self._list.append(item)
        self.trigger_callbacks()

    def remove(self, item: Any):
        self._list.remove(item)
        self.trigger_callbacks()

    def insert(self, index: int, item: Any):
        self._list.insert(index, item)
        self.trigger_callbacks()

    def pop(self, index: int = -1):
        poped_item = self._list.pop(index)
        self.trigger_callbacks()
        return poped_item

    def extend(self, iterable: Iterable):
        self._list.extend(iterable)
        self.trigger_callbacks()

    def clear(self):
        self._list = []

    def count(self, item):
        return self._list.count(item)

    def index(self, item, start=0, stop=9223372036854775807):
        return self._list.index(item, start, stop)

    def reverse(self):
        self._list.reverse()

    def sort(self, key=float, reverse=False):
        self._list.sort(key=key, reverse=reverse)

# Cell
class DatasetDescriptor(ABC):
    """Abstract base class for descriptors of datasets.
    The private name of the descriptor is the defined name with a prefix _.
    The __get__ function will call the calculate_description function if the value of the descriptor is None and then return the value else it will just return the value of the descriptor.
    The __set__ function only allows for the attribute to be set to None, which will trigger a recomputation the next time the __get__ function is called.
    When inheriting this class the function calculate_description needs to be implemented, which defines how the private value should be calculated.
    """
    def __set_name__(self, owner, name):
        owner._descriptors.append(self)
        self.private_name = '_' + name

    def __get__(self, obj, objtype=None):
        "Attribute will be recomputed if it is None else the befor computed version will be returned."
        if getattr(obj, self.private_name) is None:
            value = self.calculate_description(obj)
            setattr(obj, self.private_name, value)
        return getattr(obj, self.private_name)

    def __set__(self, obj, value):
        "Attribute can only be set to None externaly, otherwise an ValueError will be raised."
        if value is None:
            setattr(obj, self.private_name, value)
        else:
            raise ValueError("Attribute can only be set to None externaly.")

    @abstractmethod
    def calculate_description(self, obj):
        pass

# Cell
class StringDescriptor:
    """Descriptor for strings"""
    def __set_name__(self, owner, name):
        self.private_name = '_' + name

    def __get__(self, obj, objtype=None):
        return getattr(obj, self.private_name)

    def __set__(self, obj, value):
        setattr(obj, self.private_name, value)

# Cell
class GenericDataset:
    """A generic datset that has a name and description. Data is stored under the attribute base_data. The class provides a function `reset_infered_data` which can be called to reset all descriptors."""
    _descriptors = []

    name = StringDescriptor()
    description = StringDescriptor()

    def __init__(self, base_data, name: Optional[str] = None, description: Optional[str] = None):
        self.base_data = base_data
        self.name = name
        self.description = description

    def reset_infered_data(self, new_data=None):
        """Takes an argument to be compatible with callbacks."""
        for descriptor in self._descriptors:
            descriptor.__set__(self, None)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/plotting.utils.ipynb (unless otherwise specified).

__all__ = ['toggle_legend_js', 'calculate_mixing_matrix', 'get_min_and_max_dates',
           'convert_rgb_image_to_bokeh_rgb_image', 'draw_record_with_bokeh']

# Cell
import datetime
from typing import Union, Tuple, Iterable

import numpy as np
import pandas as pd

from bokeh.plotting import figure, show
from bokeh.models import CustomJS

from icevision.visualize.draw_data import draw_record

# Cell
def toggle_legend_js(figure):
    """Creates custom JS for bokeh to hide the legend with a double tap."""
    toggle_legend_js = CustomJS(
        args=dict(leg=figure.legend[0]),
        code="""if (leg.visible){
                    leg.visible = false
                } else {
                    leg.visible = true
                }"""
    )
    return toggle_legend_js

# Cell
def calculate_mixing_matrix(data: pd.DataFrame, mixing_col: str, mixing_objects: str, return_df: bool = True) -> Union[Tuple[np.ndarray, dict], pd.DataFrame]:
    """Calculates mixing matrix for the mixing_objects column where they mix in the mixing_col.
    By standard the object class mixing matrix over the images is calculated.
    Returns the mixing matrix and the mapping between label and mixing matrix index.
    If return_df is True (default) a dataframe (instead of the mixing matrix) will be returned that can be directly consumed by histogram_2d."""
    # map labels to the mixing matrix index
    mapping = {i:j for j,i in enumerate(np.sort(data[mixing_objects].unique()))}
    num_unique_mixing_objects = data[mixing_objects].nunique()
    mixing_matrix = np.zeros([num_unique_mixing_objects, num_unique_mixing_objects])
    mixing_groups = data.groupby(mixing_col)
    # iterate over each individual element with the same mixing_col to calculate the mixing based on the mixing_objects
    for group_key, group in mixing_groups:
        # handel self mixing
        for value, count in group[mixing_objects].value_counts().iteritems():
            if count > 1:
                mixing_matrix[mapping[value], mapping[value]] += 1
        # handel mixing of different objects
        permutations = np.array(np.meshgrid(group[mixing_objects].unique(), group[mixing_objects].unique())).T.reshape(-1,2)
        for permutation in permutations:
            # avoid double counting in the self mixing
            if permutation[0] != permutation[1]:
                mixing_matrix[mapping[permutation[0]], mapping[permutation[1]]] += 1

    if return_df:
        df_dict = {"values": [], "col_name": [], "row_name": []}
        for row_name, row in zip(mapping, mixing_matrix):
            df_dict["values"] += row.tolist()
            df_dict["row_name"] += [row_name]*len(mapping)
            df_dict["col_name"] += mapping
        return pd.DataFrame(df_dict)
    return mixing_matrix, mapping

# Cell
def get_min_and_max_dates(dates: Iterable[datetime.datetime]) -> Tuple[datetime.datetime, datetime.datetime]:
    """Returns the min and max date. If all dates are the same the max date is moved one day forward."""
    min_date = min(dates).replace(microsecond=0, second=0, minute=0, hour=0)
    max_date = max(dates).replace(microsecond=0, second=0, minute=0, hour=0)
    # make sure the min and max values are at least a day appart
    if min_date == max_date:
        max_date = max_date.replace(day=max_date.day+1)
    return min_date, max_date

# Cell
def convert_rgb_image_to_bokeh_rgb_image(img: np.ndarray) -> np.ndarray:
    """Convertes a image in the form of a numpy array to an array that can be shown by bokeh."""
    img = np.flipud(img)
    img = img.astype(np.uint8)
    bokeh_img = np.empty((img.shape[0],img.shape[1]), dtype=np.uint32)
    view = bokeh_img.view(dtype=np.uint8).reshape((img.shape[0],img.shape[1], 4))
    view[:,:, 0] = img[:,:,0]
    view[:,:, 1] = img[:,:,1]
    view[:,:, 2] = img[:,:,2]
    view[:,:, 3] = 255
    return bokeh_img

# Cell
def draw_record_with_bokeh(
    record,
    class_map=None,
    display_label=True,
    display_bbox=False,
    display_mask=False,
    display_keypoints=False,
    return_figure=False,
    width=None,
    height=None
):
    """Draws a record or returns a bokeh figure containing the image."""
    img = draw_record(
            record=record,
            class_map=class_map,
            display_label=display_label,
            display_bbox=display_bbox,
            display_mask=display_mask,
            display_keypoints=display_keypoints,
        )

    # create bokeh figure with the plot
    bokeh_img = convert_rgb_image_to_bokeh_rgb_image(img)

    # make sure the aspect ratio of the image is retained, if only the width of hight is given
    if width is None and height is not None:
        plot_width = int(img.shape[1]/img.shape[0] * height)
        plot_height = height
    elif height is None and width is not None:
        plot_width = width
        plot_height = int(img.shape[0]/img.shape[1] * width)
    else:
        plot_width = img.shape[1] if width is None else width
        plot_height = img.shape[0] if height is None else height

    p = figure(tools="reset, wheel_zoom, box_zoom, save, pan", width=plot_width, height=plot_height, x_range=(0, img.shape[1]), y_range=(img.shape[0], 0), x_axis_location="above")
    p.xgrid.grid_line_color = None
    p.ygrid.grid_line_color = None
    p.image_rgba([bokeh_img], x=0, y=img.shape[0], dw=img.shape[1], dh=img.shape[0], level="image")
    if return_figure:
        return p
    else:
        show(p)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/metrics.ipynb (unless otherwise specified).

__all__ = ['APObjectDetection', 'APInstanceSegmentation']

# Cell

from typing import Union, Optional, Any, Iterable, Callable
import os
import shutil
from abc import ABC, abstractmethod
from joblib import delayed, Parallel

import numpy as np

from pycocotools import mask as mask_utils

from .utils import string_to_erles

# Cell
class APObjectDetection:
    """A faster implementaiton for the (m)AP scores."""
    def __init__(self, data, ious=None):
        self.data = data
        self.ious = ious if ious is not None else np.arange(0.5, 1, 0.05).round(2)
        self.metric_data = self.get_metric_data()

    @staticmethod
    def calculate_iou(pred_box, gt_box):
        px1, py1, px2, py2 = pred_box
        tx1, ty1, tx2, ty2 = gt_box

        # return 0 if the boxes don't intersect
        if (tx2 < px1 or px2 < tx1 or ty2 < py1 or py2 < ty1):
            return 0, 0, 0, 0
        else:
            lower_x = max(tx1, px1)
            upper_x = min(tx2, px2)
            lower_y = max(ty1, py1)
            upper_y = min(ty2, py2)
            intersection_area = (upper_x-lower_x) * (upper_y-lower_y)
            gt_box_area = (tx2-tx1) * (ty2-ty1)
            pred_box_area = (px2-px1) * (py2-py1)
            iou = intersection_area / (gt_box_area + pred_box_area - intersection_area)
            return iou, pred_box_area, gt_box_area, intersection_area

    def get_image_stats(self, gt_boxes, pred_boxes, iou_threshold):
        """
        Returns: tp, fp, fn, :if additional_stats: x_center_offsets, y_center_offsets, center_distances, used_gt_box_areas_normalized, used_pred_box_areas_normalized, used_gt_box_areas_normalized, used_pred_box_areas_normalized
        """
        if pred_boxes is None:
            return 0,  0, len(gt_boxes), [], [], [], [], [], [], []
        if len(gt_boxes) == 0:
            return 0, len(pred_boxes), 0, [], [], [], [], [], [], []
        else:
            # calculate ious and log their mapping with box indices
            gt_box_indices = []
            pred_box_indices = []
            ious = []
            gt_box_areas = []
            pred_box_areas = []
            intersection_areas = []
            for pred_box_index, pred_box in enumerate(pred_boxes):
                for gt_box_index, gt_box in enumerate(gt_boxes):
                    iou, pred_box_area, gt_box_area, intersection_area = self.calculate_iou(pred_box, gt_box)
                    if iou >= iou_threshold:
                        gt_box_indices.append(gt_box_index)
                        pred_box_indices.append(pred_box_index)
                        ious.append(iou)
                        pred_box_areas.append(pred_box_area)
                        gt_box_areas.append(gt_box_area)
                        intersection_areas.append(intersection_area)

            # check if any hits happend
            if len(ious) == 0:
                return 0, len(pred_boxes), len(gt_boxes), [], [], [], [], [], [], []
            else:
                # select matches based on iou
                indices_descending = np.argsort(ious)[::-1]
                gt_match_indices = []
                pred_match_indices = []
                x_center_offsets = []
                y_center_offsets = []
                center_distances = []
                unused_gt_box_areas_normalized = []
                unused_pred_box_areas_normalized = []
                used_gt_box_areas_normalized = []
                used_pred_box_areas_normalized = []
                for index in indices_descending:
                    gt_index = gt_box_indices[index]
                    pred_index = pred_box_indices[index]
                    if (gt_index not in gt_match_indices) and (pred_index not in pred_match_indices):
                        gt_match_indices.append(gt_index)
                        pred_match_indices.append(pred_index)
                        # calculate additional stats
                        pred_box_x1, pred_box_y1, pred_box_x2, pred_box_y2 = pred_boxes[pred_box_index]
                        gt_box_x1, gt_box_y1, gt_box_x2, gt_box_y2 = gt_boxes[gt_box_index]

                        x_center_offset = ((pred_box_x1+pred_box_x2)-(gt_box_x1+gt_box_x2))/2
                        y_center_offset = ((pred_box_y1+pred_box_y2)-(gt_box_y1+gt_box_y2))/2
                        x_center_offsets.append(x_center_offset)
                        y_center_offsets.append(y_center_offset)
                        center_distances.append((x_center_offset**2+y_center_offset**2)**0.5)
                        unused_gt_box_areas_normalized.append((gt_box_areas[index]-intersection_areas[index])/gt_box_areas[index])
                        unused_pred_box_areas_normalized.append((pred_box_areas[index]-intersection_areas[index])/pred_box_areas[index])
                        used_gt_box_areas_normalized.append(intersection_areas[index]/gt_box_areas[index])
                        used_pred_box_areas_normalized.append(intersection_areas[index]/pred_box_areas[index])

                return len(gt_match_indices), len(pred_boxes) - len(pred_match_indices), len(gt_boxes) - len(gt_match_indices), x_center_offsets, y_center_offsets, center_distances, unused_gt_box_areas_normalized, unused_pred_box_areas_normalized, used_gt_box_areas_normalized, used_pred_box_areas_normalized

    def get_precision_and_recall(self, gt, pred, iou):
        """gt and pred need to be sored dicts with the lowest score being the first entry"""
        tps, fps, fns = [], [], []
        precisions, recalls, score_thresholds = [], [], []
        x_center_offsets, y_center_offsets, center_distances, unused_gt_box_areas_normalized, unused_pred_box_areas_normalized, used_gt_box_areas_normalized, used_pred_box_areas_normalized = [], [], [], [], [], [], []

        if pred is None:
            return {
                "tp": np.array([0]), "fp": [sum(len(gt_boxes) for gt_boxes in gt.values())], "fn": np.array([0]),
                "precision": np.array([0]), "recall": np.array([0]), "scores": np.array([0]),
                "ap11": 0, "ap": 0, "monotonic_recalls": np.array([0]), "monotonic_precisions": np.array([0]),
                "ap11_recalls": np.array([0]), "ap11_precisions": np.array([0]), "x_center_offsets": np.array([0]),
                "y_center_offsets": np.array([0]), "center_distances": np.array([0]),
                "unused_gt_box_areas_normalized": np.array([0]), "unused_pred_box_areas_normalized": np.array([0]),
                "used_gt_box_areas_normalized": np.array([0]), "used_pred_box_areas_normalized": np.array([0])
            }

        scores = list(pred.keys())
        pred_boxes = list(pred.values())
        # loop over scores to calculate statistics for the score
        for score_index, score in enumerate(scores):
            score_tp, score_fp, score_fn = 0, 0, 0
            score_x_center_offsets, score_y_center_offsets, score_center_distances, score_unused_gt_box_areas_normalized, score_unused_pred_box_areas_normalized = [], [], [], [], []
            score_used_gt_box_areas_normalized, score_used_pred_box_areas_normalized = [], []
            # create dict with active predicitons (prediction with the same or higher score)
            active_preds = {}
            for pred_entry in pred_boxes[score_index:]:
                for filename, bbox in zip(pred_entry["filename"], pred_entry["bboxes"]):
                    if filename not in active_preds.keys():
                        active_preds[filename] = [bbox]
                    else:
                        active_preds[filename].append(bbox)
            # loop over gt images
            for filename, image_gt_boxes in gt.items():
                img_tp, img_fp, img_fn, img_x_center_offsets, img_y_center_offsets, img_center_distances, img_unused_gt_box_areas_normalized, img_unused_pred_box_areas_normalized, img_used_gt_box_areas_normalized, img_used_pred_box_areas_normalized = self.get_image_stats(image_gt_boxes, active_preds.get(filename, None), iou)
                score_tp += img_tp
                score_fp += img_fp
                score_fn += img_fn
                score_x_center_offsets += img_x_center_offsets
                score_y_center_offsets += img_y_center_offsets
                score_center_distances += img_center_distances
                score_unused_gt_box_areas_normalized += img_unused_gt_box_areas_normalized
                score_unused_pred_box_areas_normalized += img_unused_pred_box_areas_normalized
                score_used_gt_box_areas_normalized += img_used_gt_box_areas_normalized
                score_used_pred_box_areas_normalized += img_used_pred_box_areas_normalized
            # calculate precision and recall for the threshold
            score_precision = score_tp/(score_tp + score_fp) if score_tp + score_fp > 0 else 0
            score_recall = score_tp/(score_tp + score_fn) if score_tp + score_fn > 0 else 0

            tps.append(score_tp)
            fps.append(score_fp)
            fns.append(score_fn)
            precisions.append(score_precision)
            recalls.append(score_recall)
            score_thresholds.append(score)

            x_center_offsets.append(score_x_center_offsets)
            y_center_offsets.append(score_y_center_offsets)
            center_distances.append(score_center_distances)
            unused_gt_box_areas_normalized.append(score_unused_gt_box_areas_normalized)
            unused_pred_box_areas_normalized.append(score_unused_pred_box_areas_normalized)
            used_gt_box_areas_normalized.append(score_used_gt_box_areas_normalized)
            used_pred_box_areas_normalized.append(score_used_pred_box_areas_normalized)

        # convert data to np.arrays for further processing
        tps = np.array(tps)
        fps = np.array(fps)
        fns = np.array(fns)
        precisions = np.array(precisions)
        recalls = np.array(recalls)
        score_thresholds = np.array(score_thresholds)

        # calculate additional stats

        # AP11
        precisions_at_recall_value = []
        for recall_value in np.linspace(0.0, 1.0, 11):
            indices = np.argwhere(np.array(recalls) >= recall_value).flatten()
            precision_max = max(precisions[indices]) if indices.size > 0 else 0
            precisions_at_recall_value.append(precision_max)
        ap11 = np.mean(precisions_at_recall_value)

        #AP
        sorted_indices = np.argsort(recalls)
        sorted_recalls = recalls[sorted_indices]
        sorted_precision = precisions[sorted_indices]
        # make the precision values monotonically
        calc_recalls = [0] + sorted_recalls.tolist() + [1]
        calc_precisions = [0] + sorted_precision.tolist() + [0]
        for i in range(len(calc_recalls)-2, -1, -1):
            calc_precisions[i] = max(calc_precisions[i], calc_precisions[i+1])
        # get indices where the recall value changes
        changing_index_list = []
        for i in range(1, len(calc_recalls)):
            if calc_recalls[i] != calc_recalls[i-1]:
                changing_index_list.append(i)
        ap = 0.0
        for i in changing_index_list:
            ap += ((calc_recalls[i]-calc_recalls[i-1])*calc_precisions[i])

        return {
            "tp": tps, "fp": fps, "fn": fns, "precision": precisions, "recall": recalls, "scores": score_thresholds,
            "ap11": ap11, "ap": ap, "monotonic_recalls": np.array(calc_recalls), "monotonic_precisions": np.array(calc_precisions),
            "ap11_recalls": np.linspace(0.0, 1.0, 11), "ap11_precisions": np.array(precisions_at_recall_value), "x_center_offsets": x_center_offsets,
            "y_center_offsets": y_center_offsets, "center_distances": center_distances,
            "unused_gt_box_areas_normalized": unused_gt_box_areas_normalized, "unused_pred_box_areas_normalized": unused_pred_box_areas_normalized,
            "used_gt_box_areas_normalized": used_gt_box_areas_normalized, "used_pred_box_areas_normalized": used_pred_box_areas_normalized
        }

    @staticmethod
    def prepare_data(df):
        ground_truth, preds = df[df["is_prediction"] == False].sort_values("score"), df[df["is_prediction"] == True].sort_values("score")

        pred_dict = {}
        for index, row in preds.iterrows():
            if row["label"] not in pred_dict.keys():
                pred_dict[row["label"]] = {row["score"]: {"bboxes": [[row["bbox_xmin"], row["bbox_ymin"], row["bbox_xmax"], row["bbox_ymax"]]], "filename": [row["filename"]]}}
            else:
                if not row["filename"] in pred_dict[row["label"]].keys():
                    pred_dict[row["label"]][row["score"]] = {"bboxes": [[row["bbox_xmin"], row["bbox_ymin"], row["bbox_xmax"], row["bbox_ymax"]]], "filename": [row["filename"]]}
                else:
                    pred_dict[row["label"]][row["score"]]["bboxes"].append([row["bbox_xmin"], row["bbox_ymin"], row["bbox_xmax"], row["bbox_ymax"]])
                    pred_dict[row["label"]][row["score"]]["filename"].append(row["filename"])

        gt_dict = {}
        for index, row in ground_truth.iterrows():
            if row["label"] not in gt_dict.keys():
                gt_dict[row["label"]] = {row["filename"]: [[row["bbox_xmin"], row["bbox_ymin"], row["bbox_xmax"], row["bbox_ymax"]]]}
            else:
                if not row["filename"] in gt_dict[row["label"]].keys():
                    gt_dict[row["label"]][row["filename"]] = [[row["bbox_xmin"], row["bbox_ymin"], row["bbox_xmax"], row["bbox_ymax"]]]
                else:
                    gt_dict[row["label"]][row["filename"]].append([row["bbox_xmin"], row["bbox_ymin"], row["bbox_xmax"], row["bbox_ymax"]])
        return gt_dict, pred_dict

    @staticmethod
    def filter_data(df, filter_key_word):
        if filter_key_word == "AP":
            return df
        elif filter_key_word == "AP_small":
            return df[(df["area"] < 32**2)]
        elif filter_key_word == "AP_medium":
            return df[((32**2 <= df["area"]) & (df["area"] < 96**2))]
        elif filter_key_word == "AP_large":
            return df[96**2 <= df["area"]]

    def get_metric_data(self):
        analysis_data = {}
        for analysis_type in ["AP", "AP_small", "AP_medium", "AP_large"]:
            filtered_df = self.filter_data(self.data, analysis_type)
            gt_dict, pred_dict = self.prepare_data(filtered_df)
            class_names = gt_dict.keys()
            class_data = {}
            for class_name in class_names:
                iou_data = {}
                for iou in self.ious:
                    res = self.get_precision_and_recall(gt_dict[class_name], pred_dict.get(class_name, None), iou)
                    iou_data[iou] = res
                iou_data["ap"] = np.array([iou["ap"] for iou in iou_data.values()]).mean()
                class_data[class_name] = iou_data
            class_data["map"] = np.array([class_entry["ap"] for class_entry in class_data.values()]).mean() if len(class_data.values()) > 0 else 0
            analysis_data[analysis_type] = class_data
        return analysis_data

# Cell
class APInstanceSegmentation:
    """A faster implementaiton for the (m)AP scores."""
    def __init__(self, data, ious=None):
        self.data = data
        self.ious = ious if ious is not None else np.arange(0.5, 1, 0.05).round(2)
        self.metric_data = self.get_metric_data()

    @staticmethod
    def calculate_iou(pred_mask_array, gt_mask_array):
        mask_combination = pred_mask_array + gt_mask_array
        intersection_area = (mask_combination == 2).sum()
        non_intersecting_area = (mask_combination == 1).sum()
        iou = intersection_area / (non_intersecting_area + intersection_area)
        return iou, intersection_area

    def get_image_stats(self, gt_masks, pred_masks, iou_threshold):
        """
        Returns: tp, fp, fn, :if additional_stats: x_center_offsets, y_center_offsets, center_distances, used_gt_mask_areas_normalized, used_pred_mask_areas_normalized, used_gt_mask_areas_normalized, used_pred_mask_areas_normalized
        """
        if pred_masks is None:
            return 0,  0, len(gt_masks["masks"]), [], [], [], [], [], [], []
        if len(gt_masks) == 0:
            return 0, len(pred_masks["masks"]), 0, [], [], [], [], [], [], []
        else:
            # calculate ious and log their mapping with mask indices
            gt_mask_indices = []
            pred_mask_indices = []
            ious = []
            gt_mask_areas = []
            pred_mask_areas = []
            intersection_areas = []
            for pred_mask_index, (pred_mask, pred_area) in enumerate(zip(pred_masks["masks"], pred_masks["areas"])):
                for gt_mask_index, (gt_mask, gt_area) in enumerate(zip(gt_masks["masks"], gt_masks["areas"])):
                    iou, intersection_area = self.calculate_iou(pred_mask, gt_mask)
                    if iou >= iou_threshold:
                        gt_mask_indices.append(gt_mask_index)
                        pred_mask_indices.append(pred_mask_index)
                        ious.append(iou)
                        pred_mask_areas.append(pred_area)
                        gt_mask_areas.append(gt_area)
                        intersection_areas.append(intersection_area)
            # check if any hits happend
            if len(ious) == 0:
                return 0, len(pred_masks["masks"]), len(gt_masks["masks"]), [], [], [], [], [], [], []
            else:
                # select matches based on iou
                indices_descending = np.argsort(ious)[::-1]
                gt_match_indices = []
                pred_match_indices = []
                x_center_offsets = []
                y_center_offsets = []
                center_distances = []
                unused_gt_mask_areas_normalized = []
                unused_pred_mask_areas_normalized = []
                used_gt_mask_areas_normalized = []
                used_pred_mask_areas_normalized = []
                for index in indices_descending:
                    gt_index = gt_mask_indices[index]
                    pred_index = pred_mask_indices[index]
                    if (gt_index not in gt_match_indices) and (pred_index not in pred_match_indices):
                        gt_match_indices.append(gt_index)
                        pred_match_indices.append(pred_index)
                        # calculate additional stats
                        pred_mask_array = pred_masks["masks"][pred_mask_index]
                        pred_mask_y_indices, pred_mask_x_indices = np.where(pred_mask_array == 1)
                        pred_mask_x1, pred_mask_y1, pred_mask_x2, pred_mask_y2 = pred_mask_x_indices[0], pred_mask_x_indices[-1], pred_mask_y_indices[0], pred_mask_y_indices[-1]

                        gt_mask_array = gt_masks["masks"][gt_mask_index]
                        gt_mask_y_indices, gt_mask_x_indices = np.where(gt_mask_array == 1)
                        gt_mask_x1, gt_mask_y1, gt_mask_x2, gt_mask_y2 = gt_mask_x_indices[0], gt_mask_x_indices[-1], gt_mask_y_indices[0], gt_mask_y_indices[-1]

                        x_center_offset = ((pred_mask_x1+pred_mask_x2)-(gt_mask_x1+gt_mask_x2))/2
                        y_center_offset = ((pred_mask_y1+pred_mask_y2)-(gt_mask_y1+gt_mask_y2))/2
                        x_center_offsets.append(x_center_offset)
                        y_center_offsets.append(y_center_offset)
                        center_distances.append((x_center_offset**2+y_center_offset**2)**0.5)
                        unused_gt_mask_areas_normalized.append((gt_mask_areas[index]-intersection_areas[index])/gt_mask_areas[index])
                        unused_pred_mask_areas_normalized.append((pred_mask_areas[index]-intersection_areas[index])/pred_mask_areas[index])
                        used_gt_mask_areas_normalized.append(intersection_areas[index]/gt_mask_areas[index])
                        used_pred_mask_areas_normalized.append(intersection_areas[index]/pred_mask_areas[index])

                return len(gt_match_indices), len(pred_masks["masks"]) - len(pred_match_indices), len(gt_masks["masks"]) - len(gt_match_indices), x_center_offsets, y_center_offsets, center_distances, unused_gt_mask_areas_normalized, unused_pred_mask_areas_normalized, used_gt_mask_areas_normalized, used_pred_mask_areas_normalized

    def get_precision_and_recall(self, gt, pred, iou):
        """gt and pred need to be sored dicts with the lowest score being the first entry"""
        tps, fps, fns = [], [], []
        precisions, recalls, score_thresholds = [], [], []
        x_center_offsets, y_center_offsets, center_distances, unused_gt_mask_areas_normalized, unused_pred_mask_areas_normalized, used_gt_mask_areas_normalized, used_pred_mask_areas_normalized = [], [], [], [], [], [], []

        if pred is None:
            return {
                "tp": np.array([0]), "fp": [sum(len(gt_masks) for gt_masks in gt.values())], "fn": np.array([0]),
                "precision": np.array([0]), "recall": np.array([0]), "scores": np.array([0]),
                "ap11": 0, "ap": 0, "monotonic_recalls": np.array([0]), "monotonic_precisions": np.array([0]),
                "ap11_recalls": np.array([0]), "ap11_precisions": np.array([0]), "x_center_offsets": np.array([0]),
                "y_center_offsets": np.array([0]), "center_distances": np.array([0]),
                "unused_gt_mask_areas_normalized": np.array([0]), "unused_pred_mask_areas_normalized": np.array([0]),
                "used_gt_mask_areas_normalized": np.array([0]), "used_pred_mask_areas_normalized": np.array([0])
            }, iou

        scores = list(pred.keys())
        pred_masks = list(pred.values())
        # loop over scores to calculate statistics for the score
        for score_index, score in enumerate(scores):
            score_tp, score_fp, score_fn = 0, 0, 0
            score_x_center_offsets, score_y_center_offsets, score_center_distances, score_unused_gt_mask_areas_normalized, score_unused_pred_mask_areas_normalized = [], [], [], [], []
            score_used_gt_mask_areas_normalized, score_used_pred_mask_areas_normalized = [], []
            # create dict with active predicitons (prediction with the same or higher score)
            active_preds = {"masks": [], "areas": []}
            for pred_entry in pred_masks[score_index:]:
                for filename, mask, area in zip(pred_entry["filename"], pred_entry["masks"], pred_entry["areas"]):
                    if filename not in active_preds.keys():
                        active_preds[filename] = {"masks": [mask], "areas": [area]}
                    else:
                        active_preds[filename]["masks"].append(mask)
                        active_preds[filename]["areas"].append(area)
            # loop over gt images
            for filename, image_gt_masks in gt.items():
                img_tp, img_fp, img_fn, img_x_center_offsets, img_y_center_offsets, img_center_distances, img_unused_gt_mask_areas_normalized, img_unused_pred_mask_areas_normalized, img_used_gt_mask_areas_normalized, img_used_pred_mask_areas_normalized = self.get_image_stats(image_gt_masks, active_preds.get(filename, None), iou)
                score_tp += img_tp
                score_fp += img_fp
                score_fn += img_fn
                score_x_center_offsets += img_x_center_offsets
                score_y_center_offsets += img_y_center_offsets
                score_center_distances += img_center_distances
                score_unused_gt_mask_areas_normalized += img_unused_gt_mask_areas_normalized
                score_unused_pred_mask_areas_normalized += img_unused_pred_mask_areas_normalized
                score_used_gt_mask_areas_normalized += img_used_gt_mask_areas_normalized
                score_used_pred_mask_areas_normalized += img_used_pred_mask_areas_normalized
            # calculate precision and recall for the threshold
            score_precision = score_tp/(score_tp + score_fp) if score_tp + score_fp > 0 else 0
            score_recall = score_tp/(score_tp + score_fn) if score_tp + score_fn > 0 else 0

            tps.append(score_tp)
            fps.append(score_fp)
            fns.append(score_fn)
            precisions.append(score_precision)
            recalls.append(score_recall)
            score_thresholds.append(score)

            x_center_offsets.append(score_x_center_offsets)
            y_center_offsets.append(score_y_center_offsets)
            center_distances.append(score_center_distances)
            unused_gt_mask_areas_normalized.append(score_unused_gt_mask_areas_normalized)
            unused_pred_mask_areas_normalized.append(score_unused_pred_mask_areas_normalized)
            used_gt_mask_areas_normalized.append(score_used_gt_mask_areas_normalized)
            used_pred_mask_areas_normalized.append(score_used_pred_mask_areas_normalized)

        # convert data to np.arrays for further processing
        tps = np.array(tps)
        fps = np.array(fps)
        fns = np.array(fns)
        precisions = np.array(precisions)
        recalls = np.array(recalls)
        score_thresholds = np.array(score_thresholds)

        # calculate additional stats

        # AP11
        precisions_at_recall_value = []
        for recall_value in np.linspace(0.0, 1.0, 11):
            indices = np.argwhere(np.array(recalls) >= recall_value).flatten()
            precision_max = max(precisions[indices]) if indices.size > 0 else 0
            precisions_at_recall_value.append(precision_max)
        ap11 = np.mean(precisions_at_recall_value)

        #AP
        sorted_indices = np.argsort(recalls)
        sorted_recalls = recalls[sorted_indices]
        sorted_precision = precisions[sorted_indices]
        # make the prefrom joblib import Parallel, delayedcision values monotonically
        calc_recalls = [0] + sorted_recalls.tolist() + [1]
        calc_precisions = [0] + sorted_precision.tolist() + [0]
        for i in range(len(calc_recalls)-2, -1, -1):
            calc_precisions[i] = max(calc_precisions[i], calc_precisions[i+1])
        # get indices where the recall value changes
        changing_index_list = []
        for i in range(1, len(calc_recalls)):
            if calc_recalls[i] != calc_recalls[i-1]:
                changing_index_list.append(i)
        ap = 0.0
        for i in changing_index_list:
            ap += ((calc_recalls[i]-calc_recalls[i-1])*calc_precisions[i])

        return {
            "tp": tps, "fp": fps, "fn": fns, "precision": precisions, "recall": recalls, "scores": score_thresholds,
            "ap11": ap11, "ap": ap, "monotonic_recalls": np.array(calc_recalls), "monotonic_precisions": np.array(calc_precisions),
            "ap11_recalls": np.linspace(0.0, 1.0, 11), "ap11_precisions": np.array(precisions_at_recall_value), "x_center_offsets": x_center_offsets,
            "y_center_offsets": y_center_offsets, "center_distances": center_distances,
            "unused_gt_mask_areas_normalized": unused_gt_mask_areas_normalized, "unused_pred_mask_areas_normalized": unused_pred_mask_areas_normalized,
            "used_gt_mask_areas_normalized": used_gt_mask_areas_normalized, "used_pred_mask_areas_normalized": used_pred_mask_areas_normalized
        }, iou

    @staticmethod
    def prepare_data(df):
        ground_truth, preds = df[df["is_prediction"] == False].sort_values("score"), df[df["is_prediction"] == True].sort_values("score")

        pred_dict = {}
        for index, row in preds.iterrows():
            if row["label"] not in pred_dict.keys():
                pred_dict[row["label"]] = {row["score"]: {"masks": [mask_utils.decode([string_to_erles(row["erles_corrected"])]).transpose(2, 0, 1)[0,:,:]], "filename": [row["filename"]], "areas": [row["mask_area"]]}}
            else:
                if not row["filename"] in pred_dict[row["label"]].keys():
                    pred_dict[row["label"]][row["score"]] = {"masks": [mask_utils.decode([string_to_erles(row["erles_corrected"])]).transpose(2, 0, 1)[0,:,:]], "filename": [row["filename"]], "areas": [row["mask_area"]]}
                else:
                    pred_dict[row["label"]][row["score"]]["maskes"].append(mask_utils.decode([string_to_erles(row["erles_corrected"])]).transpose(2, 0, 1)[0,:,:])
                    pred_dict[row["label"]][row["score"]]["filename"].append(row["filename"])
                    pred_dict[row["label"]][row["score"]]["areas"].append(row["mask_area"])

        gt_dict = {}
        for index, row in ground_truth.iterrows():
            if row["label"] not in gt_dict.keys():
                gt_dict[row["label"]] = {row["filename"]: {"masks": [mask_utils.decode([string_to_erles(row["erles_corrected"])]).transpose(2, 0, 1)[0,:,:]], "areas": [row["mask_area"]]}}
            else:
                if not row["filename"] in gt_dict[row["label"]].keys():
                    gt_dict[row["label"]][row["filename"]] = {"masks": [mask_utils.decode([string_to_erles(row["erles_corrected"])]).transpose(2, 0, 1)[0,:,:]], "areas": [row["mask_area"]]}
                else:
                    gt_dict[row["label"]][row["filename"]]["masks"].append(mask_utils.decode([string_to_erles(row["erles_corrected"])]).transpose(2, 0, 1)[0,:,:])
                    gt_dict[row["label"]][row["filename"]]["areas"].append(row["mask_area"])
        return gt_dict, pred_dict

    @staticmethod
    def filter_data(df, filter_key_word):
        if filter_key_word == "AP":
            return df
        elif filter_key_word == "AP_small":
            return df[(df["bbox_area"] < 32**2)]
        elif filter_key_word == "AP_medium":
            return df[((32**2 <= df["bbox_area"]) & (df["bbox_area"] < 96**2))]
        elif filter_key_word == "AP_large":
            return df[96**2 <= df["bbox_area"]]

    def get_metric_data(self):
        analysis_data = {}
        for analysis_type in ["AP", "AP_small", "AP_medium", "AP_large"]:
            filtered_df = self.filter_data(self.data, analysis_type)
            gt_dict, pred_dict = self.prepare_data(filtered_df)
            class_names = gt_dict.keys()
            class_data = {}
            for class_name in class_names:
                iou_data = {}
                results = Parallel(n_jobs=10)(delayed(self.get_precision_and_recall)(gt_dict[class_name], pred_dict.get(class_name, None), iou) for iou in self.ious)
                for res in results:
                    iou_data[res[1]] = res[0]

                # for iou in self.ious:
                #     res = self.get_precision_and_recall(gt_dict[class_name], pred_dict.get(class_name, None), iou)
                #     iou_data[res[1]] = res[0]

                iou_data["ap"] = np.array([iou["ap"] for iou in iou_data.values()]).mean()
                class_data[class_name] = iou_data
            class_data["map"] = np.array([class_entry["ap"] for class_entry in class_data.values()]).mean() if len(class_data.values()) > 0 else 0
            analysis_data[analysis_type] = class_data
        return analysis_data


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/core.dashboards.ipynb (unless otherwise specified).

__all__ = ['Dashboard', 'BaseGallery', 'RecordDastasetGallery', 'DatasetOverview', 'MultiDatasetOverview',
           'DatasetComparison', 'DatasetFilter', 'DatasetFilterWithRangeSliderAndMultiSelect',
           'DatasetFilterWithScatter', 'DatasetGenerator', 'DatasetGeneratorScatter']

# Cell
from typing import List, Union, Optional
from abc import ABC, abstractmethod
from math import floor, ceil
import os

import numpy as np
import pandas as pd
import panel as pn
import panel.widgets as pnw

from ..plotting import *
from .data import *

# Cell
class Dashboard(ABC):
    """
    Abstract base class for dashboards. Requires the functions `build_gui` and `show` to be defined.
    build_gui should define the gui elements/plots for the dashboard and show should return a panel object that can be displayed.
    In most cases build_gui will call some other functions that return gui components and combine them into one element and than register the combined element as an attribute of the class (e.g. self.gui).
    In most cases the show function than returns the gui element defined in build_gui (e.g. self.gui).
    """
    def __init__(self, width: int = 500, height: int = 500):
        """Takes width and hight as arguments and calls the `build_gui` function at the end."""
        self.width = width
        self.height = height
        self.build_gui()

    @abstractmethod
    def show(self):
        pass

    @abstractmethod
    def build_gui(self):
        pass

# Cell
class BaseGallery(Dashboard, ABC):
    """Base class for galleries that provieds the provides the basic interface for creating a gallery. Inheritied classes need to define at least two functions `get_image_by_image_id` and `get_num_entries`."""
    def __init__(self, data, sort_cols=None, width=500, height=500):
        """Data can be anyting, the handling of the data is done via the functions `get_image_by_image_id` and `get_num_entries`"""
        self.data = data
        self.sort_cols = sort_cols
        self.num_entries = self.get_num_entries()
        self.UPDATING = False
        super().__init__(width, height)

    @abstractmethod
    def get_num_entries(self):
        pass

    @abstractmethod
    def get_image_by_index(self, index):
        pass

    def build_gui(self):
        "Builds the basic gui for the gallery."
        self.btn_prev = pnw.Button(name="<", width=int(2*self.width/6))
        self.btn_next = pnw.Button(name=">", width=int(2*self.width/6))
        self.current = pnw.TextInput(value="1", width=int(self.width/6))
        self.image_count = pn.Row("/" + str(self.num_entries), width=int(self.width/6))
        self.gui_image_selection_controlls = pn.Row(self.btn_prev, self.current, self.image_count, self.btn_next, align="center", height=50)

        if self.sort_cols is not None:
            self.sorter = pnw.Select(name="Sort by", options=self.sort_cols)
            self.sorter.param.watch(self.update_sorting, "value")
            self.sort_order = pnw.CheckButtonGroup(name="Options", options=["Desc.", "Drop duplicates"])
            self.sort_order.param.watch(self.update_sorting, "value")
            self.sort_gui = pn.Row(self.sorter, self.sort_order)
        if self.sort_cols is not None:
            self.gui_controlls = pn.Column(self.sort_gui, self.gui_image_selection_controlls)
        else:
            self.gui_controlls = self.gui_image_selection_controlls

        self.btn_prev.on_click(self._previous)
        self.btn_next.on_click(self._next)
        self.current.param.watch(self._number_input, "value")

        self._image = pn.Row(self.get_image_by_index(int(self.current.value)-1), align="center")
        self.gui = pn.Column(self.gui_controlls, self.image)

    @property
    def image(self):
        return self._image

    @image.setter
    def image(self, image):
        self._image = image
        self.gui[1] = self._image

    def _next(self, _):
        """Logic for the next button"""
        index = int(self.current.value)
        if index == self.num_entries:
            index = 1
        else:
            index += 1
        self.UPDATING = True
        self.current.value = str(index)
        self.UPDATING = False
        self.image = pn.Row(self.get_image_by_index(index-1), align="center")

    def _previous(self, _):
        """Logic for the previous button"""
        index = int(self.current.value)
        if index == 1:
            index = self.num_entries
        else:
            index -= 1
        self.UPDATING = True
        self.current.value = str(index)
        self.UPDATING = False
        self.image = pn.Row(self.get_image_by_index(index-1), align="center")

    def _number_input(self, _):
        """Logic for the number input."""
        # check for update lock because an update via the next or previous will trigger an update of the number input
        if self.UPDATING:
            return
        index = int(self.current.value)
        self.image = pn.Row(self.get_image_by_index(index-1), align="center")

    def show(self):
        return self.gui

# Cell
class RecordDastasetGallery(BaseGallery):
    """Creates a gallery for a dataset with sorting options. The dataset is required to provide a function `get_image_by_image_id` which returns an panel compativle representation of an image for a given id.
    The gallery_descriptor is a string, that is the name of the descriptor to use for getting the data to create the gallery with. The descriptor need to return a pandas dataframe.
    The img_id_col is the column that identifies the unique images (e.g. filepath of image id).
    """
    def __init__(self, dataset, gallery_desciptor, img_id_col, sort_cols=None, width=500, height=500):
        self.sort_cols = sort_cols
        self.gallery_desciptor = gallery_desciptor
        self.img_id_col = img_id_col
        if sort_cols is None:
            self.index_mapping = getattr(dataset, self.gallery_desciptor).sort_values(img_id_col).drop_duplicates(self.img_id_col).reset_index(drop=True)
        else:
            self.index_mapping = getattr(dataset, self.gallery_desciptor).sort_values(sort_cols[0]).drop_duplicates(self.img_id_col).reset_index(drop=True)
        super().__init__(dataset, sort_cols, width, height)

    def get_num_entries(self):
        return getattr(self.data, self.gallery_desciptor).sort_values(self.img_id_col).drop_duplicates(self.img_id_col).shape[0]

    def get_image_by_index(self, index):
        """Converts index to image_id and calls the `get_image_by_image_id` function."""
        height_subtracor = 50 if self.sort_cols is None else 100
        return self.data.get_image_by_image_id(self.index_mapping.iloc[index][self.img_id_col], width=self.width, height=self.height-height_subtracor)

    def update_sorting(self, event):
        """Calculates the new index order for sorting."""
        sort_ascending = False if  "Desc." in self.sort_order.value else True
        data = getattr(self.data, self.gallery_desciptor).sort_values(self.sorter.value, ascending=sort_ascending)
        if "Drop duplicates" in self.sort_order.value:
            data = data.drop_duplicates(self.img_id_col)
        self.num_entries = data.shape[0]
        self.image_count = pn.Row("/" + str(self.num_entries), width=int(self.width/6))
        self.gui_image_selection_controlls[2] = self.image_count
        self.index_mapping = data.reset_index(drop=True)

# Cell
class DatasetOverview(Dashboard):
    """Provides a genric overview dashboard for datasets."""
    DESCRIPTOR_DATA = "data"
    DESCRIPTOR_STATS = "stats"

    def __init__(self, dataset: GenericDataset, height: int = 500, width: int = 500):
        """Creates an overview of a dataset"""
        self.dataset = dataset
        super().__init__(height=height, width=width)

    def _generate_dataset_tab(self):
        overview_table = table_from_dataframe(getattr(self.dataset, self.DESCRIPTOR_DATA), width=self.width, height=self.height)
        return overview_table

    def _generate_datset_stats_tab(self):
        overview_table = table_from_dataframe(getattr(self.dataset, self.DESCRIPTOR_STATS), width=self.width, height=self.height)
        return overview_table

    def build_gui(self):
        dataset_tab = self._generate_dataset_tab()
        dataset_stats_tab = self._generate_datset_stats_tab()
        self.gui = pn.Tabs(("Dataset overview", dataset_tab), ("Dataset stats overview", dataset_stats_tab), align="start")

    def show(self):
        return self.gui

# Cell
class MultiDatasetOverview(Dashboard):
    """
    Same as the `DatasetOverview` dashboard but takes a list of datasets and presents the dataset stats in a table.
    The list of datasets can be expanded by adding new entryies to the datasets attribute or elements can be removed with the delete button.
    """
    DESCRIPTOR_DATA = "stats"

    def __init__(self, datasets: Union[List[GenericDataset], ObservableList], height=100, width=1000, with_del_button=False):
        self.datasets = datasets if isinstance(datasets, ObservableList) else ObservableList(datasets)
        self.datasets.register_callback(self.update_table)
        self.with_del_button = with_del_button
        super().__init__(width=width, height=height)

    def create_overview_df(self):
        return pd.concat([getattr(dataset, self.DESCRIPTOR_DATA) for dataset in self.datasets]) if len(self.datasets) > 0 else pd.DataFrame()

    def build_gui(self):
        self.delete_button = pnw.Button(name="Delete", width=self.width, height=25)
        self.overview_table = table_from_dataframe(self.create_overview_df(), height=self.height-25)
        if self.with_del_button:
            self.delete_button.on_click(self.delete_entry)
            self.gui = pn.Column(self.delete_button, self.overview_table)
        else:
            self.gui = pn.Column(self.overview_table)

    def delete_entry(self, clicks):
        selection = self.overview_table.selection
        self.datasets.list = [dataset for index, dataset in enumerate(self.datasets) if index not in selection]

    def update_table(self, event):
        self.overview_table.value = self.create_overview_df()
        self.overview_table.height = height=self.height-25

    def show(self):
        return self.gui

# Cell
class DatasetComparison(Dashboard):
    """Same as the `DatasetOverview` Dashboard but takes a list of datasets and presents there overviews next to each other."""
    DESCRIPTOR_DATA = "data"
    DESCRIPTOR_STATS = "stats"

    def _get_descriptor_for_all_datasets(self, descriptor_name):
        return [getattr(dataset, descriptor_name) for dataset in self.datasets]

    def __init__(self, datasets: List[GenericDataset], height: int = 500, width: int = 500):
        """Creates an overview of a dataset"""
        self.datasets = datasets
        super().__init__(height=height, width=width)

    def _generate_dataset_tab(self):
        overview_table = table_from_dataframe(self._get_descriptor_for_all_datasets(self.DESCRIPTOR_DATA), width=self.width, height=self.height)
        return pn.Column(*overview_table)

    def _generate_datset_stats_tab(self):
        overview_table = table_from_dataframe(self._get_descriptor_for_all_datasets(self.DESCRIPTOR_STATS), width=self.width, height=self.height)
        return pn.Column(*overview_table)

    def build_gui(self):
        dataset_tab = self._generate_dataset_tab()
        dataset_stats_tab = self._generate_datset_stats_tab()
        self.gui = pn.Tabs(("Dataset overview", dataset_tab), ("Dataset stats overview", dataset_stats_tab), align="start")

    def show(self):
        return self.gui

# Cell
class DatasetFilter(Dashboard, ABC):
    """Abstract base class for generic filters on datasets. Allows for filtering of data using different controlls, for more on controlls see: `plotting.controlls`."""
    DESCRIPTOR_DATA = "data"

    def __init__(self, dataset: GenericDataset, columns: Optional[List[str]] = None, height: int = 500, width: int = 500, filter_width: Optional[int] = None, filter_height: Optional[int] = None, n_cols: int = None):
        self.dataset = dataset
        self.columns = columns if columns is not None else getattr(self.dataset, self.DESCRIPTOR_DATA).columns
        self.n_cols = n_cols if n_cols is not None else ceil(len(self.columns)**0.5)
        self.n_rows = (len(self.columns)//self.n_cols) + min(1, len(self.columns)%self.n_cols)
        self.filter_height = int(height/self.n_rows) if filter_height is None else filter_height
        self.filter_width = int(width/self.n_cols) if filter_width is None else filter_width
        self.filters = []
        self.UPDATING = False
        super().__init__(height=height, width=width)

    def build_gui(self):
        """All filters used below need to have two functions get_selection and update_with_mask."""
        data_selection = getattr(self.dataset, self.DESCRIPTOR_DATA)[self.columns]
        self.generate_filters(data_selection)
        # put the images in the grid
        self.gui = pn.GridSpec(ncols=self.n_cols, nrows=self.n_rows, width=self.width, height=self.height)
        for index, gui_filter in enumerate(self.filters):
            self.gui[index//self.n_cols, index%self.n_cols] = gui_filter.show()
        # hook all control elements to the update functions
        for single_filter in self.filters:
            single_filter.register_callback(self.update_plots)

    @abstractmethod
    def generate_filters(self, dataselection):
        """Write handler for the different column types of the dataframe."""
        pass

    def _update_plots(self, current_selection):
        mask = np.array(self.get_selection())
        final_mask = np.logical_or(mask, current_selection)
        for single_filter in self.filters:
            single_filter.update_with_mask(final_mask)

    def update_plots(self, event, old=None, new=None):
        if self.UPDATING:
            return
        else:
            self.UPDATING = True
            self._update_plots(event)
            self.UPDATING = False

    def show(self):
        return self.gui

    def get_selection(self):
        """Returns binary mask that combines all filter masks with an `and`"""
        mask = self.filters[0].get_selection()
        for single_filter in self.filters[1:]:
            mask = np.logical_and(mask, single_filter.get_selection())
        return mask

    def register_callback(self, callback):
        """Register callback to every underlying filter"""
        for filter in self.filters:
            filter.register_callback(callback)

# Cell
class DatasetFilterWithRangeSliderAndMultiSelect(DatasetFilter):
    def generate_filters(self, data_selection):
        # generate filters
        for column in data_selection.columns:
            if pd.api.types.is_numeric_dtype(data_selection[column]):
                self.filters.append(RangeFilter(data_selection[column].values, column, height=self.filter_height, width=self.filter_width))
            elif pd.api.types.is_categorical_dtype(data_selection[column]) or pd.api.types.is_string_dtype(data_selection[column]):
                self.filters.append(CategoricalFilter(data_selection[column].values, column, height=self.filter_height, width=self.filter_width))

# Cell
class DatasetFilterWithScatter(DatasetFilter):
    def __init__(self, dataset: GenericDataset, columns: Optional[List[str]] = None, height: int = 500, width: int = 500, filter_width: Optional[int] = None, filter_height: Optional[int] = None, n_cols: int = None):
        self.scatter_filter = None
        super().__init__(dataset, columns, height, width, filter_width, filter_height, n_cols)

    def build_gui(self):
        """All filters used below need to have two functions get_selection and update_with_mask."""
        data_selection = getattr(self.dataset, self.DESCRIPTOR_DATA)[self.columns]
        self.generate_filters(data_selection)
        # put the images in the grid
        self.categorical_grid = pn.GridSpec(ncols=self.n_cols, nrows=self.n_rows, width=self.width, height=self.height//2)
        # use all categorical filters
        for index, gui_filter in enumerate(self.filters[:-1]):
            self.categorical_grid[index//self.n_cols, index%self.n_cols] = gui_filter.show()
        # hook all control elements to the update functions
        for single_filter in self.filters:
            single_filter.register_callback(self.update_plots)
        self.gui = pn.Column(self.categorical_grid, self.scatter_filter.show())

    def generate_filters(self, data_selection):
        # generate filters
        numeric_cols = []
        for column in data_selection.columns:
            if pd.api.types.is_numeric_dtype(data_selection[column]):
                numeric_cols.append(column)
            elif pd.api.types.is_categorical_dtype(data_selection[column]) or pd.api.types.is_string_dtype(data_selection[column]):
                self.filters.append(CategoricalFilter(data_selection[column].values, column, height=self.filter_height, width=self.filter_width))
        # generate generic scatter selector
        self.scatter_filter = GenericMulitScatterFilter(data_selection[numeric_cols], height=self.height//2)
        self.filters.append(self.scatter_filter)

# Cell
class DatasetGenerator(Dashboard):
    """This dashboards can be used to generate datasets by filtering the data with filer dashboars."""
    DESCRIPTOR_DATA = "data"
    DESCRIPTOR_STATS = "stats"

    DATASET_FILTER_COLUMNS = None
    DATASET_FILTER = DatasetFilterWithRangeSliderAndMultiSelect
    MULTI_DATASET_OVERVIEW = MultiDatasetOverview
    DATASET_OVERVIEW = DatasetOverview

    def __init__(self, dataset, with_dataset_overview=True, width=500, height=500):
        self.base_dataset = dataset
        self.created_datasets = ObservableList([])
        self.with_dataset_overview = with_dataset_overview
        super().__init__(width, height)

    def build_gui(self):
        self.dataset_filter = self.DATASET_FILTER(self.base_dataset, columns=self.DATASET_FILTER_COLUMNS, width=self.width, height=self.height-50)
        self.dataset_filter_create_dataset_button = pnw.Button(name="Create", height=50)
        self.dataset_filter_create_dataset_button.on_click(self.create_dataset)
        self.dataset_filter_with_export = pn.Column(pn.Row(self.dataset_filter.show(), align="center"), pn.Row(self.dataset_filter_create_dataset_button, align="center"))

        if self.with_dataset_overview:
            self.created_datasets_overview = self.MULTI_DATASET_OVERVIEW(self.created_datasets, with_del_button=True, height=150, width=self.width)
            self.selected_dataset_overview = self.DATASET_OVERVIEW(self.base_dataset, height=self.height-250, width=self.width).gui
        else:
            self.created_datasets_overview = self.MULTI_DATASET_OVERVIEW(self.created_datasets, with_del_button=True, height=self.height-100, width=self.width)
            self.selected_dataset_overview = pn.Spacer(sizing_mode="stretch_both")
        self.export_gui = self.create_export_gui()
        self.datasets_overview = pn.Column(pn.Row(self.created_datasets_overview.gui, align="center"), pn.Row(self.selected_dataset_overview, align="center"), pn.Row(self.export_gui, align="center"))
        if self.with_dataset_overview:
            self.created_datasets_overview.overview_table.param.watch(self.update_dataset_overview, "selection")

        self.gui = pn.Tabs(("Dataset Filter", self.dataset_filter_with_export), ("Dataset Overview", self.datasets_overview))

    def create_export_gui(self):
        self.export_path = pnw.TextInput(name="Export path", value="datasets", height=50)
        self.export_button = pnw.Button(name="Export", align="end", height=50)
        self.export_button.on_click(self.export_datasets)

        export_dataset_name = "" if len(self.created_datasets) == 0 else self.created_datasets[self.export_dataset_overview.selection[0]].name
        export_description_name = "" if len(self.created_datasets) == 0 else self.created_datasets[self.export_dataset_overview.selection[0]].description
        self.export_name_input = pnw.TextInput(name="Dataset name", value=export_dataset_name, height=50)
        self.export_name_input.param.watch(self.change_dataset_name, "value")
        self.export_description_input = pnw.TextAreaInput(name="Description", value=export_description_name, height=50)
        self.export_description_input.param.watch(self.change_dataset_description, "value")
        return pn.Column(pn.Row(self.export_name_input, self.export_description_input), pn.Row(self.export_path, self.export_button))

    def change_dataset_name(self, event):
        index = self.created_datasets_overview.overview_table.selection[0]
        self.created_datasets[index].name = event.new
        self.created_datasets_overview.update_table(event)

    def change_dataset_description(self, event):
        index = self.created_datasets_overview.overview_table.selection[0]
        self.created_datasets[index].description = event.new
        self.created_datasets_overview.update_table(event)

    def update_dataset_overview(self, event):
        self.selected_dataset_overview = self.DATASET_OVERVIEW(self.created_datasets[event.new[0]], height=self.height-350)
        self.datasets_overview[1] = pn.Row(self.selected_dataset_overview.show())
        self.export_name_input.value = self.created_datasets[event.new[0]].name
        self.export_description_input.value = self.created_datasets[event.new[0]].description

    def create_dataset(self, clicks):
        mask = self.dataset_filter.get_selection()
        new_sub_dataset = self.base_dataset.create_new_from_mask(self.base_dataset, mask)
        self.created_datasets.append(new_sub_dataset)

    def export_datasets(self, clicks):
        export_path = self.export_path.value
        if not os.path.isdir(export_path):
            os.makedirs(export_path)
        for dataset in self.created_datasets:
            dataset.save(export_path)

    def show(self):
        return self.gui

# Cell
class DatasetGeneratorScatter(DatasetGenerator):
    """Same as `DatasetGenerator` but replaces the range slider + hists for the numerical columns with a sactter plot selection."""
    DATASET_FILTER = DatasetFilterWithScatter


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/data.ipynb (unless otherwise specified).

__all__ = ['RecordDataframeParser', 'RecordDataset', 'ResultsDataset', 'BboxRecordDataframeParser',
           'DataDescriptorBbox', 'StatsDescriptorBbox', 'ImageStatsDescriptorBbox', 'ClassStatsDescriptorBbox',
           'GalleryStatsDescriptorBbox', 'BboxRecordDataset', 'PrecisionRecallMetricsDescriptorObjectDetection',
           'ObjectDetectionResultsDataset', 'InstanceSegmentationRecordDataframeParser',
           'DataDescriptorInstanceSegmentation', 'StatsDescriptorInstanceSegmentation',
           'ImageStatsDescriptorInstanceSegmentation', 'ClassStatsDescriptorInstanceSegmentation',
           'GalleryStatsDescriptorInstanceSegmentation', 'InstanceSegmentationRecordDataset',
           'PrecisionRecallMetricsDescriptorInstanceSegmentation', 'InstanceSegmentationResultsDataset']

# Cell
import datetime
from typing import Union, Optional, List
import os
import json
from copy import deepcopy
from random import shuffle
from abc import ABC

import numpy as np
import pandas as pd
from PIL import Image
import panel as pn

import icevision
from icevision.core.record import BaseRecord
from icevision.core.record_defaults import ObjectDetectionRecord, InstanceSegmentationRecord
import icevision.parsers as parsers
from icevision.data.data_splitter import SingleSplitSplitter
from icevision.core.bbox import BBox
from icevision.core.mask import EncodedRLEs, MaskArray
from icevision.core.class_map import ClassMap

from pycocotools import mask as mask_utils

from fastprogress import master_bar, progress_bar

from .plotting.utils import draw_record_with_bokeh
from .metrics import APObjectDetection, APInstanceSegmentation
from .core.data import *
from .utils import erles_to_counts_to_utf8, erles_to_string, string_to_erles, correct_mask, decorrect_mask

# Cell
class RecordDataframeParser(parsers.Parser):
    """IceVision parser for pandas dataframes. This parser is mostly used by the RecordDataset to load records from a saved RecordDataset."""
    def __init__(self, record_template):
        super().__init__(record_template)

    def __iter__(self):
        for group in self.record_dataframe.groupby("filepath"):
            yield group[1]

    def __len__(self):
        return self.record_dataframe["filepath"].nunique

    def record_id(self, o):
        return o.iloc[0]["id"]

    def parse_fields(self, o, record, is_new):
        width, height = o.iloc[0]["width"], o.iloc[0]["height"]
        record.set_filepath(o.iloc[0]["filepath"])
        record.set_img_size((width, height))
        record.detection.set_class_map(self.class_map)
        record.detection.add_labels(o["label"])

# Cell
class RecordDataset(GenericDataset, ABC):
    """Base class dashboard datasets that are based on IceVision records."""
    def __init__(self, records: Union[List[BaseRecord], ObservableList, str], class_map, name=None, description=None):
        if isinstance(records, str):
            self.load_from_file(records)
        else:
            if isinstance(records, icevision.data.record_collection.RecordCollection):
                records = records._records._list
            self.records = records if isinstance(records, ObservableList) else ObservableList(records)
            self.class_map = class_map
        super().__init__(self.records, name=name, description=description)
        self.records.register_callback(self.reset_infered_data)

    def __repr__(self):
        base_string = ""
        if getattr(self, "stats_dataset", None) is not None:
            for col in self.stats_dataset.columns:
                base_string += str(col) + ": " + str(self.stats_dataset[col][0]) + " | "
            base_string = base_string[:-2]
        return base_string

    def __getitem__(self, index):
        return self.records[index]

    def __len__(self):
        return len(self.records)

    def __add__(self, other):
        """Add two RecordDatases together but sets name and description to None for the new RecordDataset."""
        if sorted(self.class_map._id2class) != sorted(other.class_map._id2class):
            raise ValueError("The two RecordDatasets don't have the same class_map")
        else:
            return RecordDataset(self.records+other.records, class_map=self.class_map)

    def split_in_train_and_val(self, train_fraction):
        records = list(self.records)
        indices = list(range(len(records)))
        shuffle(indices)
        if train_fraction > 1:
            train_fraction /= len(records)
        train = [records[index] for index in indices[:int(len(records)*train_fraction)]]
        val = [records[index] for index in indices[int(len(records)*train_fraction):]]
        return train, val

    @property
    def num_images(self):
        return len(self)

    @classmethod
    def load_from_record_dataframe(cls, record_data_df: pd.DataFrame, class_map=None, name=None, description=None):
        if class_map is None:
            class_map = ClassMap(json.loads(record_data_df["class_map"].values[0])[1:])
        records = cls.parse_df_to_records(record_data_df, class_map)
        return cls(records, class_map=class_map, name=name, description=description)

    @staticmethod
    def parse_df_to_records(record_data_df):
        raise NotImplementedError()

    def load_from_file(self, path):
        data = json.load(open(path))
        df = pd.DataFrame(data["data"])
        self.class_map = ClassMap(data["class_map"])
        self._name = data["name"]
        self._description = data["description"]
        records = self.parse_df_to_records(df, self.class_map)
        self.records = ObservableList(records)

    def save(self, save_path):
        if not os.path.isdir(save_path):
            os.makedirs(save_path, exist_ok=True)
        save_name = "dataset" if self.name == "" or self.name is None else self.name
        if not os.path.isfile(os.path.join(save_path, save_name+".json")):
            save_name = save_name+".json"
        else:
            counter = 1
            while True:
                save_name = save_name+"("+str(counter)+").json"
                if os.path.isfile(os.path.join(save_path, save_name)):
                    counter += 1
                else:
                    break

        class_map = self.class_map if self.class_map is not None else self.create_class_map_from_record_df(self.data)
        save_data = {"name": self.name, "description": self.description, "data": self.data.to_dict(), "class_map": class_map._id2class}

        json.dump(save_data, open(os.path.join(save_path, save_name), "w"), default=str)

    def get_image_by_index(self, index, width, height):
        return draw_record_with_bokeh(self[index], class_map=self.class_map, width=None, height=height, return_figure=True)

    @classmethod
    def create_new_from_mask(cls, cls_instance, mask):
        selection = cls_instance.data[mask]
        filepaths = np.unique(selection["filepath"]).tolist()
        new_records = [record for record in cls_instance.records if str(record.filepath) in filepaths]
        return cls(new_records, cls_instance.class_map)

# Cell
class ResultsDataset(GenericDataset):
    """Dashboard dataset for the results of and object detection system."""
    metric_data_ap = None
    df_parser = None

    def __init__(self, dataframe, name=None, description=None):
        super().__init__(dataframe, name, description)
        # instanciate metric data and preload it
        self.metric_data_ap = None
        self.class_map = ClassMap([str(i) for i in self.base_data[["label", "label_num"]].drop_duplicates().sort_values("label_num")["label"].tolist()])

    def save(self, path):
        if not os.path.exists(os.path.join(*path.split("/")[:-1])):
            os.makedirs(os.path.join(*path.split("/")[:-1]))
        self.base_data.to_csv(path)

    def get_image_by_image_id(self, image_id, width=None, height=None):
        """For gallery dashboards"""
        df_pred = self.base_data[(self.base_data["filepath"] == image_id) & (self.base_data["is_prediction"] == True)]
        df_gt = self.base_data[(self.base_data["filepath"] == image_id) & (self.base_data["is_prediction"] == False)]

        parser_pred = self.df_parser(df_pred, self.class_map)
        parser_gt = self.df_parser(df_gt, self.class_map)
        res_gt = parser_gt.parse(show_pbar=False, autofix=False)[1][0]
        res_pred = parser_pred.parse(show_pbar=False, autofix=False)[1]
        if len(res_pred) == 0:
            res_pred = deepcopy(res_gt)
            res_pred.labels = []
            res_pred.bboxes = []
        else:
            res_pred = res_pred[0]
        plot_gt = draw_record_with_bokeh(res_gt, display_bbox=True, display_mask=True, return_figure=True, width=width, height=height)
        plot_pred = draw_record_with_bokeh(res_pred, display_bbox=True, display_mask=True, return_figure=True, width=width, height=height)
        return pn.Row(pn.Column(pn.Row("<b>Ground Truth</b>",  align="center"), plot_gt), pn.Column(pn.Row("<b>Prediction</b>",  align="center"), plot_pred))

    @classmethod
    def load(cls, path):
        df = pd.read_csv(path)
        return cls(df, None, None)

# Cell
class BboxRecordDataframeParser(RecordDataframeParser):
    """Extends the RecordDataframeParser for object detection"""
    def __init__(self, record_dataframe, class_map):
        super().__init__(ObjectDetectionRecord())
        self.record_dataframe = record_dataframe
        self.class_map = class_map

    def parse_fields(self, o, record, is_new):
        super().parse_fields(o, record, is_new)
        record.detection.add_bboxes([BBox(annot[1]["bbox_xmin"], annot[1]["bbox_ymin"], annot[1]["bbox_xmax"], annot[1]["bbox_ymax"]) for annot in o.iterrows()])

# Cell
class DataDescriptorBbox(DatasetDescriptor):
    """Dashboard dataset for object detection"""
    def calculate_description(self, obj):
        """Aggregates stats from a list of records and returns a pandas dataframe with the aggregated stats. The creation time is not necessarily the real creation time.
        This depends on the OS, for more information see: https://docs.python.org/3/library/os.html#os.stat_result."""
        data = []
        for index,record in enumerate(obj.records):
            record_commons, record_detections = record.as_dict()["common"], record.aggregate_objects()["detection"]
            for label, bbox in zip(record_detections["labels"], record_detections["bboxes"]):
                file_stats = record.filepath.stat()
                bbox_width = bbox["bbox_width"]
                bbox_height = bbox["bbox_height"]
                area = bbox_width*bbox_height
                area_normalized = area / (record.width * record.height)
                bbox_ratio = bbox_width / bbox_height
                data.append(
                    {
                        "id": record_commons["record_id"], "width": record.width, "height": record.height, "label": label, "area_square_root": bbox["bbox_sqrt_area"], "area_square_root_normalized": area_normalized**0.5,
                        "bbox_xmin": bbox["bbox_x"], "bbox_xmax": bbox["bbox_x"]+bbox["bbox_width"], "bbox_ymin": bbox["bbox_y"], "bbox_ymax": bbox["bbox_y"]+bbox["bbox_height"],
                        "bbox_xmin_normalized": bbox["bbox_x"]/record.width, "bbox_xmax_normalized": (bbox["bbox_x"]+bbox["bbox_width"])/record.width, "bbox_ymin_normalized": bbox["bbox_y"]/record.height,
                        "bbox_ymax_normalized": (bbox["bbox_y"]+bbox["bbox_height"])/record.height, "area": area, "area_normalized": area_normalized, "bbox_ratio": bbox_ratio, "bbox_ratio_normalized": bbox_ratio*(record.height/record.width),
                        "record_index": index, "bbox_width": bbox_width, "bbox_height": bbox_height, "bbox_width_normalized": bbox_width/record.width, "bbox_height_normalized": bbox_height/record.height,
                        "filepath": str(record.filepath), "creation_date": datetime.datetime.fromtimestamp(file_stats.st_ctime), "class_map": json.dumps(record.detection.class_map._id2class),
                        "modification_date": datetime.datetime.fromtimestamp(file_stats.st_mtime), "num_annotations": len(record_detections["bboxes"])
                    }
                )
        data = pd.DataFrame(data)
        data["label_num"] = data["label"]
        if obj.class_map is not None:
            data["label"] = data["label"].apply(obj.class_map.get_by_id)
        return data

# Cell
class StatsDescriptorBbox(DatasetDescriptor):
    def calculate_description(self, obj):
        stats_dict = {}
        stats_dict["no_imgs"] = [obj.data["filepath"].nunique()]
        stats_dict["no_classes"] = [obj.data["label"].nunique()]
        stats_dict["classes"] = [list(obj.data["label"].unique())]
        stats_dict["area_min"] = [obj.data["area"].min()]
        stats_dict["area_max"] = [obj.data["area"].max()]
        stats_dict["num_annotations_min"] = [obj.data["num_annotations"].min()]
        stats_dict["num_annotations_max"] = [obj.data["num_annotations"].max()]
        stats_dict["name"] = [obj._name]
        stats_dict["description"] = [obj._description]
        return pd.DataFrame(stats_dict)

# Cell
class ImageStatsDescriptorBbox(DatasetDescriptor):
    def calculate_description(self, obj):
        """Creates a dataframe containing stats about the images."""
        stats_dict = {}
        stats_dict["Num. imgs."] = obj.data["filepath"].nunique()
        stats_dict["Min Num. Objects"] = obj.data["num_annotations"].min()
        stats_dict["Max Num. Objects"] = obj.data["num_annotations"].max()
        stats_dict["Avg. Objects/Img"] = round(obj.data["num_annotations"].mean(),2)
        df = pd.DataFrame.from_dict(stats_dict, orient="index").T
        return df

# Cell
class ClassStatsDescriptorBbox(DatasetDescriptor):
    def calculate_description(self, obj):
        """Creates a dataframe containing stats about the object classes."""
        stats_dict = {}
        label_group = obj.data.groupby("label")
        for label, group in label_group:
            label_stats = {}
            label_stats["imgs"] = group["filepath"].nunique()
            label_stats["objects"] = group.shape[0]
            label_stats["avg_objects_per_img"] = label_stats["objects"]/label_stats["imgs"]
            label_stats["frac_of_labels"] = round(label_stats["objects"]/obj.data.shape[0], 2)
            stats_dict[label] = label_stats
        df = pd.DataFrame(stats_dict).T
        df = df.rename_axis('Class').reset_index()
        return df

# Cell
class GalleryStatsDescriptorBbox(DatasetDescriptor):
    def calculate_description(self, obj):
        """Creates a dataframe containing the data for a gallery."""
        df = obj.data[["id", "area", "num_annotations", "label", "bbox_ratio", "bbox_width", "bbox_height", "width", "height"]].drop_duplicates().reset_index(drop=True)
        return df

# Cell
class BboxRecordDataset(RecordDataset):
    data = DataDescriptorBbox()
    gallery_data = GalleryStatsDescriptorBbox()
    stats_dataset = StatsDescriptorBbox()
    stats_image = ImageStatsDescriptorBbox()
    stats_class = ClassStatsDescriptorBbox()
    stats = StatsDescriptorBbox()

    def __init__(self, records: Union[List[BaseRecord], ObservableList, str], class_map=None, name=None, description=None):
        super().__init__(records, class_map, name, description)
        self.record_index_image_id_map = {str(record.filepath): index for index, record in enumerate(self.records)}
        self.data = None
        self.gallery_data = None
        self.stats_dataset = None
        self.stats_image = None
        self.stats_class = None
        self.stats = None

    @staticmethod
    def parse_df_to_records(record_data_df, class_map):
        return BboxRecordDataframeParser(record_data_df, class_map).parse(SingleSplitSplitter())[0]

    def get_image_by_image_id(self, image_id, width, height):
        index = self.record_index_image_id_map[image_id]
        return draw_record_with_bokeh(self[index], display_bbox=True, display_mask=True, class_map=self.class_map, width=None, height=height, return_figure=True)

# Cell
class PrecisionRecallMetricsDescriptorObjectDetection(DatasetDescriptor):
    def __init__(self, ious=None):
        if ious is None:
            self.ious = np.arange(0.5, 1, 0.05).round(2)
        else:
            self.ious = ious

    def calculate_description(self, obj):
        return APObjectDetection(obj.base_data, self.ious).metric_data

# Cell
class ObjectDetectionResultsDataset(ResultsDataset):
    """Dashboard dataset for the results of and object detection system."""
    metric_data_ap = PrecisionRecallMetricsDescriptorObjectDetection()
    df_parser = BboxRecordDataframeParser

    @classmethod
    def init_from_preds_and_samples(cls, predictions, samples_plus_losses, padded_along_shortest=True, class_map=None, name=None, description=None):
        """The input_records are required because they are the only information source with the image stats(width, etc.) for the image on the disk."""
        data = []
        for index, (prediction, sample_plus_loss) in enumerate(zip(predictions, samples_plus_losses)):
            # TODO: At the moment only resize_and_pad or resize are handeled. Check if there are other edge cases that need to be included
            # The correction requires that the sample_plus_loss has the scaled image sizes (not the padded ones or the original ones)
            # correct the width and height to the values of the original image
            prediction = prediction.pred.as_dict()["detection"]

            if len(prediction["labels"]) > 0:
                losses = sample_plus_loss.losses
                sample_plus_loss_dict = sample_plus_loss.as_dict()
                # bbox correction
                img = Image.open(str(sample_plus_loss.common.filepath))
                width, height = img.size[0], img.size[1]
                # use bool to int for padded_along_shortest and int(sample_plus_loss["width"] < sample_plus_loss["height"]) to avoid if branches
                resize_factor = sample_plus_loss_dict["common"]["width"]/width if width > height else sample_plus_loss_dict["common"]["height"]/height
                resized_x = resize_factor * width
                resized_y = resize_factor * height
                pad_x = sample_plus_loss_dict["common"]["width"]-resized_x
                pad_y = sample_plus_loss_dict["common"]["height"]-resized_y
                correct_x = lambda x: (x-pad_x/2)/resize_factor
                correct_y = lambda y: (y-pad_y/2)/resize_factor

                for label, bbox, score in zip(prediction["labels"], prediction["bboxes"], prediction["scores"]):
                    xmin, xmax, ymin, ymax = correct_x(bbox.xmin), correct_x(bbox.xmax), correct_y(bbox.ymin), correct_y(bbox.ymax)
                    file_stats = sample_plus_loss.common.filepath.stat()
                    bbox_width = xmax - xmin
                    bbox_height = ymax - ymin
                    area = bbox_width * bbox_height
                    area_normalized = area / (width * height)
                    bbox_ratio = bbox_width / bbox_height
                    image_data = {
                            "id": sample_plus_loss_dict["common"]["record_id"], "width": width, "height": height, "label": label, "area_square_root": area**2, "area_square_root_normalized": area_normalized**2,
                            "score": score, "bbox_xmin": xmin, "bbox_xmax": xmax, "bbox_ymin": ymin, "bbox_ymax": ymax, "area": area,
                            "bbox_xmin_normalized": xmin/width, "bbox_xmax_normalized": xmax/width, "bbox_ymin_normalized": ymin/height, "bbox_ymax_normalized": ymax/height,
                            "area_normalized": area_normalized, "bbox_ratio": bbox_ratio, "bbox_ratio_normalized": bbox_ratio*(height/width), "record_index": index,
                            "bbox_width": bbox_width, "bbox_height": bbox_height, "bbox_width_normalized": bbox_width/width, "bbox_height_normalized": bbox_height/height,
                            "filepath": str(sample_plus_loss.common.filepath), "filename": str(sample_plus_loss.common.filepath).split("/")[-1], "creation_date": datetime.datetime.fromtimestamp(file_stats.st_ctime),
                            "modification_date": datetime.datetime.fromtimestamp(file_stats.st_mtime), "num_annotations": len(prediction["bboxes"]), "is_prediction": True,
                        }
                    for key, value in losses.items():
                        image_data[key] = value
                    data.append(image_data)

            if len(sample_plus_loss_dict["detection"]["labels"]) > 0:
                for label, bbox in zip(sample_plus_loss_dict["detection"]["labels"], sample_plus_loss_dict["detection"]["bboxes"]):
                    xmin, xmax, ymin, ymax = correct_x(bbox.xmin), correct_x(bbox.xmax), correct_y(bbox.ymin), correct_y(bbox.ymax)
                    file_stats = sample_plus_loss.common.filepath.stat()
                    bbox_width = xmax - xmin
                    bbox_height = ymax - ymin
                    area = bbox_width * bbox_height
                    area_normalized = area / (width * height)
                    bbox_ratio = bbox_width / bbox_height
                    image_data = {
                            "id": sample_plus_loss_dict["common"]["record_id"], "width": width, "height": height, "label": label,
                            "score": 999, "bbox_xmin": xmin, "bbox_xmax": xmax, "bbox_ymin": ymin, "bbox_ymax": ymax, "area": area, "area_square_root": area**2, "area_square_root_normalized": area_normalized**2,
                            "area_normalized": area_normalized, "bbox_ratio": bbox_ratio, "record_index": index,
                            "bbox_xmin_normalized": xmin/width, "bbox_xmax_normalized": xmax/width, "bbox_ymin_normalized": ymin/height, "bbox_ymax_normalized": ymax/height,
                            "bbox_width": bbox_width, "bbox_height": bbox_height,  "bbox_width_normalized": bbox_width/width, "bbox_height_normalized": bbox_height/height,
                            "filepath": str(sample_plus_loss.common.filepath), "filename": str(sample_plus_loss.common.filepath).split("/")[-1], "creation_date": datetime.datetime.fromtimestamp(file_stats.st_ctime),
                            "modification_date": datetime.datetime.fromtimestamp(file_stats.st_mtime), "num_annotations": len(prediction["bboxes"]), "is_prediction": False,
                        }
                    for key, value in losses.items():
                        image_data[key] = value
                    data.append(image_data)

        data = pd.DataFrame(data)

        data["label_num"] = data["label"]
        if class_map is not None:
            data["label_num"] = data["label"].apply(class_map.get_by_name)

        if not any(data["is_prediction"] == True):
            raise ValueError("No predictions found.")

        return cls(data, name, description)

# Cell
class InstanceSegmentationRecordDataframeParser(RecordDataframeParser):
    """Extends the RecordDataframeParser for instance segmentation"""
    def __init__(self, record_dataframe, class_map):
        super().__init__(InstanceSegmentationRecord())
        self.record_dataframe = record_dataframe
        self.class_map = class_map

    def parse_fields(self, o, record, is_new):
        super().parse_fields(o, record, is_new)
        bboxes = []
        masks  = []
        for annot in o.iterrows():
            bboxes.append(BBox(annot[1]["bbox_xmin"], annot[1]["bbox_ymin"], annot[1]["bbox_xmax"], annot[1]["bbox_ymax"]))
            masks.append(EncodedRLEs([string_to_erles(annot[1]["erles_corrected"])]))
        record.detection.add_bboxes(bboxes)
        record.detection.add_masks(masks)

# Cell
class DataDescriptorInstanceSegmentation(DatasetDescriptor):
    """Dashboard dataset for object detection"""
    def calculate_description(self, obj):
        """Aggregates stats from a list of records and returns a pandas dataframe with the aggregated stats. The creation time is not necessarily the real creation time.
        This depends on the OS, for more information see: https://docs.python.org/3/library/os.html#os.stat_result."""
        data = []
        summe = 0
        for index,record in enumerate(obj.records):
            record_commons, record_detections = record.as_dict()["common"], record.as_dict()["detection"]

            if isinstance(record_detections["masks"][0], EncodedRLEs):
                masks_to_iterate_over = [entry.erles[0] for entry in record_detections["masks"]]
            else:
                masks_to_iterate_over = record_detections["masks"][0].to_erles(None,None).erles

            for label, bbox, mask in zip(record_detections["labels"], record_detections["bboxes"], masks_to_iterate_over):
                mask_array = mask_utils.decode([mask])
                file_stats = record.filepath.stat()
                area = bbox.width*bbox.height
                area_normalized = area / (record.width * record.height)
                bbox_ratio = bbox.width / bbox.height

                data.append(
                    {
                        "id": record_commons["record_id"], "width": record.width, "height": record.height, "label": label, "bbox_area_square_root": bbox.area**0.5, "bbox_area_square_root_normalized": area_normalized**0.5,
                        "bbox_xmin": bbox.xmin, "bbox_xmax": bbox.xmin+bbox.width, "bbox_ymin": bbox.ymin, "bbox_ymax": bbox.ymin+bbox.height,
                        "bbox_xmin_normalized": bbox.xmin/record.width, "bbox_xmax_normalized": (bbox.xmin+bbox.width)/record.width, "bbox_ymin_normalized": bbox.ymin/record.height,
                        "bbox_ymax_normalized": (bbox.ymin+bbox.height)/record.height, "bbox_area": area, "bbox_area_normalized": area_normalized, "bbox_ratio": bbox_ratio, "bbox_ratio_normalized": bbox_ratio*(record.height/record.width),
                        "record_index": index, "bbox_width": bbox.width, "bbox_height": bbox.height, "bbox_width_normalized": bbox.width/record.width, "bbox_height_normalized": bbox.height/record.height,
                        "filepath": str(record.filepath), "creation_date": datetime.datetime.fromtimestamp(file_stats.st_ctime),
                        "modification_date": datetime.datetime.fromtimestamp(file_stats.st_mtime), "num_annotations": len(record_detections["bboxes"]),
                        "erles_corrected": erles_to_string(mask), "erles": erles_to_string(mask), "mask_area": mask_array.sum(), "mask_area_normalized": mask_array.sum()/mask_array.size, "mask_area_normalized_by_bbox_area": mask_array.sum()/area,
                    }
                )

        data = pd.DataFrame(data)
        data["label_num"] = data["label"]
        if obj.class_map is not None:
            data["label_num"] = data["label"].apply(obj.class_map.get_by_name)
        return data

# Cell
class StatsDescriptorInstanceSegmentation(DatasetDescriptor):
    def calculate_description(self, obj):
        stats_dict = {}
        stats_dict["no_imgs"] = [obj.data["filepath"].nunique()]
        stats_dict["no_classes"] = [obj.data["label"].nunique()]
        stats_dict["classes"] = [list(obj.data["label"].unique())]
        stats_dict["mask_area_min"] = [obj.data["mask_area"].min()]
        stats_dict["mask_area_max"] = [obj.data["mask_area"].max()]
        stats_dict["num_annotations_min"] = [obj.data["num_annotations"].min()]
        stats_dict["num_annotations_max"] = [obj.data["num_annotations"].max()]
        stats_dict["name"] = [obj._name]
        stats_dict["description"] = [obj._description]
        return pd.DataFrame(stats_dict)

# Cell
class ImageStatsDescriptorInstanceSegmentation(ImageStatsDescriptorBbox):
    pass

# Cell
class ClassStatsDescriptorInstanceSegmentation(ClassStatsDescriptorBbox):
    def calculate_description(self, obj):
        """Creates a dataframe containing stats about the object classes."""
        stats_dict = {}
        label_group = obj.data.groupby("label")
        for label, group in label_group:
            label_stats = {}
            label_stats["imgs"] = group["filepath"].nunique()
            label_stats["objects"] = group.shape[0]
            label_stats["avg_objects_per_img"] = label_stats["objects"]/label_stats["imgs"]
            label_stats["frac_of_labels"] = round(label_stats["objects"]/obj.data.shape[0], 2)
            stats_dict[label] = label_stats
        df = pd.DataFrame(stats_dict).T
        df = df.rename_axis('Class').reset_index()
        return df

# Cell
class GalleryStatsDescriptorInstanceSegmentation(DatasetDescriptor):
    def calculate_description(self, obj):
        """Creates a dataframe containing the data for a gallery."""
        df = obj.data[["id", "bbox_area", "num_annotations", "label", "bbox_ratio", "bbox_width", "bbox_height", "mask_area", "mask_area_normalized", "width", "height"]].drop_duplicates().reset_index(drop=True)
        return df

# Cell
class InstanceSegmentationRecordDataset(RecordDataset):
    data = DataDescriptorInstanceSegmentation()
    gallery_data = GalleryStatsDescriptorInstanceSegmentation()
    stats_dataset = StatsDescriptorInstanceSegmentation()
    stats_image = ImageStatsDescriptorInstanceSegmentation()
    stats_class = ClassStatsDescriptorInstanceSegmentation()
    stats = StatsDescriptorInstanceSegmentation()

    def __init__(self, records: Union[List[BaseRecord], ObservableList, str], class_map=None, name=None, description=None):
        super().__init__(records, class_map, name, description)
        self.record_index_image_id_map = {str(record.filepath): index for index, record in enumerate(self.records)}
        self.data = None
        self.gallery_data = None
        self.stats_dataset = None
        self.stats_image = None
        self.stats_class = None
        self.stats = None

    @staticmethod
    def parse_df_to_records(record_data_df, class_map):
        return InstanceSegmentationRecordDataframeParser(record_data_df, class_map).parse(SingleSplitSplitter())[0]

    def get_image_by_image_id(self, image_id, width, height):
        index = self.record_index_image_id_map[image_id]
        return draw_record_with_bokeh(self[index], display_bbox=True, display_mask=True, class_map=self.class_map, width=None, height=height, return_figure=True)

# Cell
class PrecisionRecallMetricsDescriptorInstanceSegmentation(DatasetDescriptor):
    def __init__(self, ious=None):
        if ious is None:
            self.ious = np.arange(0.5, 1, 0.05).round(2)
        else:
            self.ious = ious

    def calculate_description(self, obj):
        return APInstanceSegmentation(obj.base_data, self.ious).metric_data

# Cell
class InstanceSegmentationResultsDataset(ResultsDataset):
    """Dashboard dataset for the results of and object detection system."""
    metric_data_ap = PrecisionRecallMetricsDescriptorInstanceSegmentation()
    df_parser = InstanceSegmentationRecordDataframeParser

    @staticmethod
    def get_masks_to_iterate_over(prediction):
        if len(prediction["masks"]) > 0:
            if isinstance(prediction["masks"][0], EncodedRLEs):
                masks_to_iterate_over = [entry.erles[0] for entry in prediction["masks"]]
            elif isinstance(prediction["masks"][0], MaskArray):
                masks_to_iterate_over = [entry.to_erles(None, None).erles[0] for entry in prediction["masks"]]
            else:
                masks_to_iterate_over = prediction["masks"][0].to_erles(None,None).erles
        else:
            masks_to_iterate_over = prediction["mask_array"].to_erles(None,None).erles
        return masks_to_iterate_over

    @classmethod
    def init_from_preds_and_samples(cls, predictions, samples_plus_losses, padded_along_shortest=True, class_map=None, name=None, description=None):
        """The input_records are required because they are the only information source with the image stats(width, etc.) for the image on the disk."""
        data = []
        for index, (prediction, sample_plus_loss) in enumerate(zip(predictions, samples_plus_losses)):
            # TODO: At the moment only resize_and_pad or resize are handeled. Check if there are other edge cases that need to be included
            # The correction requires that the sample_plus_loss has the scaled image sizes (not the padded ones or the original ones)
            # correct the width and height to the values of the original image
            prediction = prediction.pred.as_dict()["detection"]
            losses = sample_plus_loss.losses
            sample_plus_loss_dict = sample_plus_loss.as_dict()
            # bbox correction
            img = Image.open(sample_plus_loss.common.filepath)
            width, height = img.size[0], img.size[1]
            # use bool to int for padded_along_shortest and int(sample_plus_loss["width"] < sample_plus_loss["height"]) to avoid if branches
            resize_factor = sample_plus_loss_dict["common"]["width"]/width if width > height else sample_plus_loss_dict["common"]["height"]/height
            resized_x = resize_factor * width
            resized_y = resize_factor * height
            pad_x = sample_plus_loss_dict["common"]["width"]-resized_x
            pad_y = sample_plus_loss_dict["common"]["height"]-resized_y
            correct_x = lambda x: (x-pad_x/2)/resize_factor
            correct_y = lambda y: (y-pad_y/2)/resize_factor

            if len(prediction["labels"]) > 0:
                masks_to_iterate_over = cls.get_masks_to_iterate_over(prediction)

                for label, bbox, score, mask in zip(prediction["labels"], prediction["bboxes"], prediction["scores"], masks_to_iterate_over):
                    mask_array = mask_utils.decode([mask])
                    xmin, xmax, ymin, ymax = correct_x(bbox.xmin), correct_x(bbox.xmax), correct_y(bbox.ymin), correct_y(bbox.ymax)
                    file_stats = sample_plus_loss.common.filepath.stat()
                    bbox_width = xmax - xmin
                    bbox_height = ymax - ymin
                    area = bbox_width * bbox_height
                    area_normalized = area / (width * height)
                    bbox_ratio = bbox_width / bbox_height
                    # correct mask
                    corrected_mask = correct_mask(mask_array, pad_x, pad_y, width, height)
                    corrected_mask = corrected_mask.to_erles(None, None).erles[0]

                    image_data = {
                        "id": sample_plus_loss_dict["common"]["record_id"], "width": width, "height": height, "label": label, "bbox_area_square_root": area**2, "bbox_area_square_root_normalized": area_normalized**2,
                        "score": score, "bbox_xmin": xmin, "bbox_xmax": xmax, "bbox_ymin": ymin, "bbox_ymax": ymax, "bbox_area": area,
                        "bbox_xmin_normalized": xmin/width, "bbox_xmax_normalized": xmax/width, "bbox_ymin_normalized": ymin/height, "bbox_ymax_normalized": ymax/height,
                        "bbox_area_normalized": area_normalized, "bbox_ratio": bbox_ratio, "bbox_ratio_normalized": bbox_ratio*(height/width), "record_index": index,
                        "bbox_width": bbox_width, "bbox_height": bbox_height, "bbox_width_normalized": bbox_width/width, "bbox_height_normalized": bbox_height/height,
                        "erles_corrected": erles_to_string(corrected_mask), "erles": erles_to_string(mask), "mask_area": mask_array.sum(), "mask_area_normalized": mask_array.sum()/mask_array.size, "mask_area_normalized_by_bbox_area": mask_array.sum()/area,
                        "filepath": str(sample_plus_loss.common.filepath), "filename": str(sample_plus_loss.common.filepath).split("/")[-1], "creation_date": datetime.datetime.fromtimestamp(file_stats.st_ctime),
                        "modification_date": datetime.datetime.fromtimestamp(file_stats.st_mtime), "num_annotations": len(prediction["bboxes"]), "is_prediction": True,
                    }
                    for key, value in losses.items():
                        image_data[key] = value
                    data.append(image_data)

            if len(sample_plus_loss_dict["detection"]["labels"]) > 0:
                masks_to_iterate_over = cls.get_masks_to_iterate_over(sample_plus_loss_dict["detection"])
                for label, bbox, mask in zip(sample_plus_loss_dict["detection"]["labels"], sample_plus_loss_dict["detection"]["bboxes"], masks_to_iterate_over):
                    mask_array = mask_utils.decode([mask])
                    xmin, xmax, ymin, ymax = correct_x(bbox.xmin), correct_x(bbox.xmax), correct_y(bbox.ymin), correct_y(bbox.ymax)
                    file_stats = sample_plus_loss.common.filepath.stat()
                    bbox_width = xmax - xmin
                    bbox_height = ymax - ymin
                    area = bbox_width * bbox_height
                    area_normalized = area / (width * height)
                    bbox_ratio = bbox_width / bbox_height
                    # correct mask
                    decorrected_mask = decorrect_mask(mask_array, int(pad_x/2), int(pad_y/2), width, height)
                    decorrected_mask = decorrected_mask.to_erles(None, None).erles[0]

                    # Below the erles_corrected is set to the mask and not the mask_corrected because atm icevision does not transform the masks in the record. Only in the output of the data loader is transformed
                    image_data = {
                        "id": sample_plus_loss_dict["common"]["record_id"], "width": width, "height": height, "label": label,
                        "score": 999, "bbox_xmin": xmin, "bbox_xmax": xmax, "bbox_ymin": ymin, "bbox_ymax": ymax, "bbox_area": area, "bbox_area_square_root": area**2, "bbox_area_square_root_normalized": area_normalized**2,
                        "bbox_area_normalized": area_normalized, "bbox_ratio": bbox_ratio, "record_index": index,
                        "bbox_xmin_normalized": xmin/width, "bbox_xmax_normalized": xmax/width, "bbox_ymin_normalized": ymin/height, "bbox_ymax_normalized": ymax/height,
                        "bbox_width": bbox_width, "bbox_height": bbox_height,  "bbox_width_normalized": bbox_width/width, "bbox_height_normalized": bbox_height/height,
                        "erles_corrected": erles_to_string(mask), "erles": erles_to_string(decorrected_mask), "mask_area": mask_array.sum(), "mask_area_normalized": mask_array.sum()/mask_array.size, "mask_area_normalized_by_bbox_area": mask_array.sum()/area,
                        "filepath": str(sample_plus_loss.common.filepath), "filename": str(sample_plus_loss.common.filepath).split("/")[-1], "creation_date": datetime.datetime.fromtimestamp(file_stats.st_ctime),
                        "modification_date": datetime.datetime.fromtimestamp(file_stats.st_mtime), "num_annotations": len(prediction["bboxes"]), "is_prediction": False,
                    }
                    for key, value in losses.items():
                        image_data[key] = value
                    data.append(image_data)

        data = pd.DataFrame(data)
        data["label_num"] = data["label"]
        if class_map is not None:
            data["label_num"] = data["label"].apply(class_map.get_by_name)

        if not any(data["is_prediction"] == True):
            raise ValueError("No predictions found.")

        return cls(data, name, description)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/plotting.controls.ipynb (unless otherwise specified).

__all__ = ['Filter', 'RangeFilter', 'CategoricalFilter', 'TimeFilter', 'ScatterFilter', 'GenericMulitScatterFilter']

# Cell
import datetime
from typing import Any, Iterable, Union, Tuple, List, Optional
from abc import ABCMeta, abstractmethod
from inspect import signature

import numpy as np
import pandas as pd
from bokeh.plotting import figure, show
from bokeh.io import push_notebook
from bokeh.models import ColumnDataSource, LassoSelectTool
import panel as pn
import panel.widgets as pnw


from .core import *
from .utils import *

# Cell
class Filter(metaclass=ABCMeta):
    """Abstract base class for filters."""
    def __init__(self, data, width=500, height=500) -> None:
        self.data = data
        self.width = width
        self.height = height
        self.build_gui()

    def mask_callback(self, callback):
        def new_callback(event):
            return callback(self.get_selection())
        return new_callback

    @abstractmethod
    def register_callback(self, callback):
        pass

    @abstractmethod
    def build_gui(self):
        pass

    @abstractmethod
    def get_selection(self, inverted):
        pass

    @abstractmethod
    def show(self):
        pass

# Cell
class RangeFilter(Filter):
    def __init__(self, data: np.ndarray, name: str, bins: int = 20, steps: int = 50, with_hist: bool = True, width: int = 500, height: int = 500) -> None:
        self.steps = steps
        self.step_width = (data.max()-data.min())/steps
        self.bins = bins
        self.name = name
        self.with_hist = with_hist
        super().__init__(data=data, width=width, height=height)

    def build_gui(self) -> None:
        val_min = min(self.data)
        val_max = max(self.data)
        # subtract and add a bit to the min and max value to ensure the whole range is captured
        dist = val_max-val_min if val_max != val_min else 1
        val_min = val_min-0.01*dist
        val_max = val_max+0.01*dist
        self.slider = pnw.RangeSlider(name=self.name, start=val_min, end=val_max, step=round(((val_max-val_min)/self.steps), 1), width=int(0.97*self.width))
        self.slider.param.watch(self.update_self, "value_throttled")
        if self.with_hist:
            self.hist = histogram(self.data, bins=self.bins, range=(self.data.min(), self.data.max()), height=100, width=self.width, remove_tools=True)
        else:
            self.hist = None
        self.gui = pn.Column(self.slider, self.hist)

    def update_self(self, event):
        mask = self.get_selection()
        self.update_with_mask(mask)

    def get_selection(self, inverted: bool = False) -> np.ndarray:
        selection = (self.data >= self.gui[0].value[0]) & (self.data <= self.gui[0].value[1])
        if inverted:
            selection = ~selection
        return selection

    def update_with_mask(self, mask: np.ndarray) -> None:
        if self.hist is None:
            return
        else:
            selection = self.data[np.logical_and(self.get_selection(), mask)]
            self.hist = histogram(selection, bins=self.bins, range=(self.data.min(), self.data.max()), height=100, width=self.width, remove_tools=True)
            self.gui[1] = self.hist

    def register_callback(self, callback):
        callback = self.mask_callback(callback)
        self.slider.param.watch(callback, "value_throttled")

    def show(self) -> pn.layout.base.Column:
        return self.gui

# Cell
class CategoricalFilter(Filter):
    """A multi select based filter for categorical data"""
    def __init__(self, data: np.ndarray, name: str, width: int = 500, height: int = 500) -> None:
        self.name = name
        super().__init__(data=data, width=width, height=height)

    def build_gui(self) -> None:
        self.selector = pnw.MultiSelect(name=self.name, options=np.unique(self.data).tolist(), value=np.unique(self.data).tolist(), width=self.width, height=self.height)
        self.gui = pn.Column(self.selector)

    def get_selection(self, inverted: bool = False) -> np.ndarray:
        selection = np.in1d(self.data, self.selector.value)
        if inverted:
            selection = ~selection
        return selection

    def show(self) -> pn.layout.base.Column:
        return self.gui

    def register_callback(self, callback):
        callback = self.mask_callback(callback)
        self.selector.param.watch(callback, "value")

    def update_with_mask(self, mask: np.ndarray, disabel_callbacks=True):
        """This does not work yet, maybe the whole structure needs to be changed to be based on the param lib."""
        self.selector.value = list(set([element for element, mask_value in zip(self.data, mask) if mask_value]))

# Cell
class TimeFilter(Filter):
    """A filter for time data with a start and an end date"""
    def __init__(self, start_times: Iterable[datetime.datetime], end_times: Iterable[datetime.datetime], start_time_label: str = "Start", end_time_label: str = "End", width: int = 500, height: int = 500) -> None:
        self.start_times = start_times
        self.end_times = end_times
        self.start_time_label = start_time_label
        self.end_time_label = end_time_label
        super().__init__((start_times, end_times), width=width, height=height)

    def build_gui(self):
        self.plot = time_arc_plot(self.start_times, self.end_times, width=self.width, height=self.height-100)
        min_start_time, max_start_time = get_min_and_max_dates(self.start_times)
        min_end_time, max_end_time = get_min_and_max_dates(self.end_times)
        min_date = min(min_start_time, min_end_time)
        max_date = max(max_start_time, max_end_time)
        self.start_time_slider = pnw.DateRangeSlider(name=self.start_time_label, start=min_date, end=max_date, width=self.width, height=50)
        self.end_time_slider = pnw.DateRangeSlider(name=self.end_time_label, start=min_date, end=max_date, width=self.width, height=50)
        self.gui = pn.Column(self.plot, self.start_time_slider, self.end_time_slider)

    def get_selection(self, inverted=False):
        start_selection = (self.start_times > self.start_time_slider.value[0]) & (self.start_times < self.start_time_slider.value[1])
        end_selection = (self.end_times > self.end_time_slider.value[0]) & (self.end_times < self.end_time_slider.value[1])
        return start_selection & end_selection

    def register_callback(self, callback):
        callback = self.mask_callback(callback)
        self.start_time_slider.param.watch(callback, "value_throttled")
        self.end_time_slider.param.watch(callback, "value_throttled")

    def show(self):
        return self.gui

# Cell
class ScatterFilter(Filter):
    """A filter based on a scatter plot with a lasso selection."""
    def __init__(self, x, y, x_label: str = "", y_label: str = "", width: int = 500, height: int = 500):
        self.x_label = x_label
        self.y_label = y_label
        super().__init__((x,y), width=width, height=height)

    def build_gui(self):
        self.source = ColumnDataSource({"x": self.data[0], "y": self.data[1]})
        p = figure(x_axis_label=self.x_label, y_axis_label=self.y_label, width=self.width, height=self.height, tools="lasso_select")
        p.scatter("x", "y", source=self.source)
        p.select_one(LassoSelectTool).select_every_mousemove = False
        self.figure = p
        self.gui = pn.Row(self.figure)

    def get_selection(self):
        selected_indices = self.source.selected.indices
        selection = [False if i not in selected_indices else True for i in range(len(self.data[0]))]
        return selection

    def mask_callback(self, callback):
        def new_callback(attr, new, old):
            callback(self.get_selection())
        return new_callback

    def register_callback(self, callback):
        callback = self.mask_callback(callback)
        self.source.selected.on_change("indices", callback)

    def show(self):
        return self.gui

# Cell
class GenericMulitScatterFilter(Filter):
    """A generic filter base on the scatter plot filter, that provides additional inputs for column selection and how the selections over the different columns should be combined."""
    def __init__(self, data, columns: Optional[List[str]] = None, mode: str = "symmetric", width: int = 500, height: int = 500):
        """mode: symmetric (default) or singular. If symmetric for a specific x-y combination the corrosponding y-x combination will be updated with the same values. Else they have different selections."""
        self.mode = mode
        self.columns = columns if columns is not None else data.columns
        self.source = ColumnDataSource(data)
        self.source.selected.on_change("indices", self.update_selection)
        self.selections = {x_key: {y_key: [False]*data.shape[0] for y_key in self.columns} for x_key in self.columns}
        super().__init__(data, width, height)

    def update_selection(self, attr, old, new):
        self.selections[self.x_select.value][self.y_select.value] = [True if index in new else False for index in range(len(self.selections[self.x_select.value][self.y_select.value]))]
        if self.mode == "symmetric":
            self.selections[self.y_select.value][self.x_select.value] = [True if index in new else False for index in range(len(self.selections[self.x_select.value][self.y_select.value]))]
        self.source.data["colors"] = self.get_colors()

    def update_plot(self, event):
        self.gui[3] = self.scatter_plot()

    def build_gui(self):
        self.x_select = pnw.Select(name="x-Axis", options=list(self.data.columns), value=self.data.columns[0], width=self.width//2, height=50)
        self.x_select.param.watch(self.update_plot, "value")
        second_select_index = 1 if len(self.data.columns) > 1 else 0
        self.y_select = pnw.Select(name="y-Axis", options=list(self.data.columns), value=self.data.columns[second_select_index], width=self.width//2, height=50)
        self.y_select.param.watch(self.update_plot, "value")
        self.combine_selections = pnw.RadioBoxGroup(name="Combine selections", options=["None", "Or", "And"], value="None", width=self.width//3, inline=True)
        self.combine_selections.param.watch(self.update_plot, "value")
        self.ignore_empty_selections = pnw.CheckBoxGroup(name="Ignore empty selections", value=["Ignore empty selections"], options=["Ignore empty selections"])
        self.ignore_empty_selections.param.watch(self.update_plot, "value")
        p = self.scatter_plot()
        legend = pn.Row("<p style='color:#B22222';>Selected in current</p>", "<p style='color:#FF00FF';>Selected in another</p>", "<p style='color:#808080';>Not selected</p>")
        self.gui = pn.Column(
            pn.Row(self.x_select, self.y_select),
            pn.Row("Combine selections with: ", self.combine_selections, self.ignore_empty_selections),
#             pn.Row("Ignore empty selections", ),
            legend, p
        )

    def scatter_plot(self):
        p = figure(x_axis_label=self.x_select.value, y_axis_label=self.y_select.value, width=self.width, height=self.height-50, tools="lasso_select")
        # get selections
        if self.combine_selections.value == "None":
            combined_selection_selected = self.selections[self.x_select.value][self.y_select.value]
        else:
            # put all singel selections into one list so they can be iterated over with zip
            if "Ignore empty selections" in self.ignore_empty_selections.value:
                single_selections = [selection for layer_1 in self.selections.values() for selection in layer_1.values() if any(selection)]
                if len(single_selections) == 0:
                    single_selections = [selection for selection in single_selections if any(selection)]
            else:
                single_selections = [selection for layer_1 in self.selections.values() for selection in layer_1.values()]
            if self.combine_selections.value == "And":
                combined_selection_selected = [all(point_selections) for point_selections in zip(*single_selections)]
            else:
                combined_selection_selected = [any(point_selections) for point_selections in zip(*single_selections)]
        self.source.selected.indices = [index for index, value in enumerate(combined_selection_selected) if value]
        self.source.data["colors"] = self.get_colors()
        radius = (self.data[self.x_select.value].max()-self.data[self.x_select.value].min() + self.data[self.y_select.value].max()-self.data[self.y_select.value].min())* 0.01
        p.scatter(self.x_select.value, self.y_select.value, source=self.source, fill_color="colors", line_color=None, radius=radius)
        p.select_one(LassoSelectTool).select_every_mousemove = False
        return p

    def get_colors(self):
        # create colors from selections
        colors = ["gray"]*self.data.shape[0]
        current_selections =self.selections[self.x_select.value][self.y_select.value]
        single_selections = [selection for layer_1 in self.selections.values() for selection in layer_1.values()]
        any_selections = [any(point_selections) for point_selections in zip(*single_selections)]
        for index, (any_selection, current_selection) in enumerate(zip(any_selections, current_selections)):
            if any_selection is True and current_selection is False:
                colors[index] = "fuchsia"
            elif current_selection is True:
                colors[index] = "firebrick"
        return colors

    def get_selection(self):
        if self.combine_selections.value == "None":
            selection = [value for value in self.selections[self.x_select.value][self.y_select.value]]
        else:
            if "Ignore empty selections" in self.ignore_empty_selections.value:
                single_selections = [selection for layer_1 in self.selections.values() for selection in layer_1.values() if any(selection)]
                if len(single_selections) == 0:
                    single_selections = [selection for selection in single_selections if any(selection)]
            else:
                single_selections = [selection for layer_1 in self.selections.values() for selection in layer_1.values()]
            if self.combine_selections.value == "And":
                selection = [all(point_selections) for point_selections in zip(*single_selections)]
            else:
                selection = [any(point_selections) for point_selections in zip(*single_selections)]
        return selection

    def mask_callback(self, callback):
        def new_callback(attr, old, new):
            return callback(self.get_selection())
        return new_callback

    def register_callback(self, callback):
        callback = self.mask_callback(callback)
        self.source.selected.on_change("indices", callback)

    def update_with_mask(self, mask):
        pass

    def show(self):
        return self.gui


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/plotting.core.ipynb (unless otherwise specified).

__all__ = ['barplot', 'histogram', 'heatmap', 'time_arc_plot', 'table_from_dataframe', 'stacked_hist',
           'categorical_2d_histogram', 'categorical_2d_histogram_with_gui', 'scatter_plot_with_gui', 'plots_as_matrix']

# Cell
import datetime
from typing import List, Union, Iterable, Tuple, Optional

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype
import bokeh
from bokeh.plotting import figure, show
from bokeh.models import ColumnDataSource, TableColumn, DataTable, LinearColorMapper, ColorBar
from bokeh.models.widgets import HTMLTemplateFormatter
from bokeh.palettes import Viridis, viridis, Category20
from bokeh.transform import factor_cmap
from bokeh.layouts import gridplot
from bokeh import events
import panel as pn
import panel.widgets as pnw

from .utils import *

# Cell
def barplot(counts: Union[np.ndarray, List[np.ndarray]], values: Union[np.ndarray, List[np.ndarray]], bar_type: Union["horizontal", "vertical"] = "horizontal", linked_axis=True, width: int = 500, height: int = 500, **kwargs) -> bokeh.plotting.Figure:
    """Creates a figure with a barplot, were the counts is the bar height and values are the labels for the bars.
    Input can be a numpy array or a list of numpy array to create multiple plots.
    """
    if isinstance(counts, list) and isinstance(values, list):
        plot_list = []
        for counts_element, values_element in zip(counts, values):
            values_element = [str(entry) for entry in values_element]
            if bar_type == "horizontal":
                if len(plot_list) == 0:
                    p = figure(width=width, height=height, y_range=values_element)
                    p.hbar(y=values_element, left=0, right=counts_element, height=0.9)
                else:
                    if linked_axis:
                        p = figure(width=width, height=height, y_range=plot_list[0].y_range, x_range=plot_list[0].x_range)
                    else:
                        p = figure(width=width, height=height, y_range=values_element)
                    p.hbar(y=values_element, left=0, right=counts_element, height=0.9)
            elif bar_type == "vertical":
                if len(plot_list) == 0:
                    p = figure(width=width, height=height, x_range=values_element)
                    p.vbar(x=values_element, bottom=0, top=counts_element, width=0.9)
                else:
                    if linked_axis:
                        p = figure(width=width, height=height, x_range=plot_list[0].x_range, y_range=plot_list[0].y_range)
                    else:
                        p = figure(width=width, height=height, x_range=values_element)
                    p.vbar(x=values_element, bottom=0, top=counts_element, width=0.9)
            else:
                raise ValueError("hist_type has to be of 'horizontal' or 'vertical'")
            plot_list.append(p)
        return plot_list
    elif isinstance(counts, np.ndarray) and isinstance(values, np.ndarray):
        values = [str(entry) for entry in values]
        if bar_type == "horizontal":
            p = figure(width=width, height=height, y_range=values)
            p.hbar(y=values, left=0, right=counts, height=0.9)
        elif bar_type == "vertical":
            p = figure(width=width, height=height, x_range=values)
            p.vbar(x=values, bottom=0, top=counts, width=0.9)
        else:
            raise ValueError("hist_type has to be of 'horizontal' or 'vertical'")
        return p
    else:
        raise TypeError("counts and values need to be of the same type (list or np.ndarray)")

# Cell
def histogram(values: Union[np.ndarray, List[np.ndarray]], bins: int = 10, range: Tuple[int] = None, density: bool = False, remove_tools: bool = False, linked_axis=True, title="", x_label="", y_label="", orientation="horizontal", width: int = 500, height: int = 500) -> bokeh.plotting.Figure:
    "Creates a histogram."
    if isinstance(values, np.ndarray):
        if orientation == "vertical":
            x_label, y_label = y_label, x_label
        p = figure(width=width, height=height, title=title, x_axis_label=x_label, y_axis_label=y_label)
        counts, edges = np.histogram(values, bins=bins, range=range, density=density)
        if orientation == "horizontal":
            p.quad(top=counts, bottom=0, left=edges[:-1], right=edges[1:])
        elif orientation == "vertical":
            p.quad(top=edges[1:], bottom=edges[:-1], left=0, right=counts)
        if remove_tools:
            p.toolbar.logo = None
            p.toolbar_location = None
        return p
    elif isinstance(values, list):
        if not isinstance(title, list):
            title = [title]*len(values)
        if not isinstance(x_label, list):
            x_labels = [x_label]*len(values)
        if not isinstance(y_label, list):
            y_labels = [y_label]*len(values)
        if orientation == "vertical":
            x_labels, y_labels = y_labels, x_labels
        plot_list = []
        for values_element, plot_title, x_label, y_label in zip(values, title, x_labels, y_labels):
            if len(plot_list) == 0 or linked_axis == False:
                p = figure(width=width, height=height, title=plot_title, x_axis_label=x_label, y_axis_label=y_label)
            else:
                p = figure(width=width, height=height, x_range=plot_list[0].x_range, y_range=plot_list[0].y_range, title=plot_title, x_axis_label=x_label, y_axis_label=y_label)
            counts, edges = np.histogram(values_element, bins=bins, range=range, density=density)
            if orientation == "horizontal":
                p.quad(top=counts, bottom=0, left=edges[:-1], right=edges[1:])
            elif orientation == "vertical":
                p.quad(top=edges[1:], bottom=edges[:-1], left=0, right=counts)
            if remove_tools:
                p.toolbar.logo = None
                p.toolbar_location = None
            plot_list.append(p)
        return plot_list
    else:
        raise TypeError("values has to be of type list or np.ndarray")

# Cell
def heatmap(data: Union[pd.DataFrame, List[pd.DataFrame]], col_x: str, col_y: str, col_values: str, color_mapper: Optional[bokeh.models.ColorMapper] = None, normalize: str = "None", link_plots: bool = True, height: int = 500, width: int = 500) -> bokeh.plotting.Figure:
    def _heatmap(df: pd.DataFrame, col_x: str, col_y: str, col_values: str, color_mapper: Optional[bokeh.models.ColorMapper] = None, normalize: str = "None", height: int = 500, width: int = 500, x_range=None, y_range=None):
        df = df.copy()

        # normalize values
        if normalize == "Column":
            for name in df[col_x].unique():
                df.loc[(df[col_x] == name, col_values)] /= df[df[col_x] == name][col_values].sum()
        elif normalize == "Row":
            for name in df[col_y].unique():
                df.loc[(df[col_y] == name, col_values)] /= df[df[col_y] == name][col_values].sum()

        # ensure the x and y column are in a categorical format
        if df[col_x].dtype != str or df[col_y].dtype != str:
            df = df[[col_x, col_y, col_values]].copy()
            df[col_x] = df[col_x].astype(str)
            df[col_y] = df[col_y].astype(str)

        if x_range is not None and y_range is not None:
            p = figure(x_range=x_range, y_range=y_range, x_axis_location="above", tools="hover", toolbar_location=None, tooltips=[('', '@'+col_values)], width=width, height=height)
        elif x_range is not None and y_range is None:
            p = figure(x_range=x_range, y_range=sorted(df[col_y].unique()), x_axis_location="above", tools="hover", toolbar_location=None, tooltips=[('', '@'+col_values)], width=width, height=height)
        elif x_range is None and y_range is not None:
            p = figure(x_range=sorted(df[col_x].unique())[::-1], y_range=y_range, x_axis_location="above", tools="hover", toolbar_location=None, tooltips=[('', '@'+col_values)], width=width, height=height)
        else:
            p = figure(x_range=sorted(df[col_x].unique())[::-1], y_range=sorted(df[col_y].unique()), x_axis_location="above", tools="hover", toolbar_location=None, tooltips=[('', '@'+col_values)], width=width, height=height)

        p.grid.grid_line_color = None
        p.axis.axis_line_color = None
        p.axis.major_tick_line_color = None
        p.axis.major_label_standoff = 0

        p.rect(
            x=col_x, y=col_y, width=1, height=1, source=df,
            fill_color={'field': col_values, 'transform': color_mapper},
            line_color=None
        )
        return p

    def _create_colorbar_and_color_mapper(min_val, max_val, color_mapper):
        # create color bar
        if color_mapper is None:
            color_mapper = LinearColorMapper(palette=Viridis[256], low=min_val, high=max_val)

        color_bar = ColorBar(
            color_mapper=color_mapper, major_label_text_font_size="7px",
            label_standoff=6, border_line_color=None, location=(0, 0)
        )
        return color_bar, color_mapper

    # handel multiple inputs here
    if isinstance(data, list):
        plot_list = []
        if link_plots:
            min_value = min(df[col_values].min() for df in data)
            max_value = max(df[col_values].max() for df in data)
            color_bar, color_mapper = _create_colorbar_and_color_mapper(min_value, max_value, color_mapper)
        for df in data:
            if not link_plots:
                color_bar, color_mapper = _create_colorbar_and_color_mapper(df[col_values].min(), df[col_values].max(), None)
            temp_heatmap = _heatmap(df, col_x, col_y, col_values, color_mapper, normalize, height, width)
            if not link_plots:
                temp_heatmap.add_layout(color_bar, 'right')
            plot_list.append(temp_heatmap)
        if link_plots:
            c_bar_figure = figure(width=65, height=height, toolbar_location=None, min_border=0, outline_line_color=None)
            c_bar_figure.add_layout(color_bar, 'right')
            plot_list.append(c_bar_figure)
        return plot_list
    elif isinstance(data, pd.DataFrame):
        color_bar, color_mapper = _create_colorbar_and_color_mapper(data[col_values].min(), data[col_values].max(), color_mapper)
        p = _heatmap(data, col_x, col_y, col_values, color_mapper, normalize, height, width)
        p.add_layout(color_bar, 'right')
        return p
    else:
        raise TypeError("Data has to be of type list or pd.Dataframe")

# Cell
def time_arc_plot(start_dates: pd.core.series.Series, end_dates: Iterable[datetime.datetime], width: int = 500, height: int  = 300) -> bokeh.plotting.Figure:
    """Creates a plot of arcs where each ark spans from a start date to an end date."""
    radius = (end_dates-start_dates)/2
    x = start_dates + radius
    p = figure(x_axis_label="Date", x_axis_type='datetime', y_range=(0, radius.max()), x_range=(start_dates.min(), end_dates.max()), width=width, height=height, tools="")
    p.yaxis.major_tick_line_color = None
    p.yaxis.minor_tick_line_color = None
    p.yaxis.major_label_text_font_size = '0pt'
    p.toolbar.logo = None
    p.toolbar_location = None
    p.arc(x=x, y=0, radius=radius, start_angle=0, end_angle=np.pi)
    return p

# Cell
def table_from_dataframe(data: Union[pd.DataFrame, List[pd.DataFrame]], columns: List[str] = None, width: int = 500, height: int = None, index_position: int = None, **kwargs) -> pn.widgets.tables.DataFrame:
    if height is None:
        upper_bound = data.shape[0] if isinstance(data, pd.DataFrame) else data[0].shape[0]
        alternative_bound = data.shape[0] if isinstance(data, pd.DataFrame) else data[0].shape[0]
        height = 25*6 if upper_bound > 5 else 25*(alternative_bound+1)

    def _create_table(df, columns, width, height):
        if columns is None:
            columns = df.columns
        selection = df[columns]
        template = """<span href="#" data-toggle="tooltip" title="<%= value %>"><%= value %></span>"""
        table = pnw.DataFrame(selection, formatters={key: HTMLTemplateFormatter(template=template) for key in selection.columns}, selection=[0], width=width, height=height, **kwargs)
        return table

    if isinstance(data, pd.DataFrame):
        return _create_table(data, columns, width, height)
    elif isinstance(data, list):
        table_list = [_create_table(df, columns, width, height) for df in data]
        return table_list
    else:
        raise TypeError("data has to be of type list or pd.DataFrame")

# Cell
def stacked_hist(data: Union[list, pd.DataFrame], x_col: str, stack_col: str, x_label: str = "", link_plots=True, width: int = 500, height: int = 500):
    """If normalize is True the biggest col is set to 1 and all other cols are scaled accordingly."""
    def _stacked_hist(dataframe, x_col, stack_col, x_label, width, height, x_range=None, y_range=None):
        data = {"x": sorted(dataframe[x_col].unique().astype(str))}
        for stack_value in dataframe[stack_col].unique():
            x_values = dataframe.groupby(stack_col).get_group(stack_value).groupby(x_col).count()
            data[str(stack_value)] = [float(x_values[x_values.index == i][stack_col].values) if i in x_values.index else 0 for i in sorted(dataframe[x_col].unique())]
        colors =  Category20[dataframe[stack_col].nunique()] if (2 < dataframe[stack_col].nunique() < 21) else viridis(dataframe[stack_col].nunique())
        legend = [stack_col+": "+i for i in dataframe[stack_col].unique().astype(str).tolist()]

        if x_range is None and y_range is None:
            p = figure(x_range=sorted(dataframe[x_col].unique().astype(str)), height=height, width=width, toolbar_location=None, tools="", tooltips="$name: @$name")
        elif x_range is None and y_range is not None:
            p = figure(x_range=sorted(dataframe[x_col].unique().astype(str)), y_range=y_range, height=height, width=width, toolbar_location=None, tools="", tooltips="$name: @$name")
        elif x_range is not None and y_range is None:
            p = figure(x_range=x_range, height=height, width=width, toolbar_location=None, tools="", tooltips="$name: @$name")
        else:
            p = figure(x_range=x_range, y_range=y_range, height=height, width=width, toolbar_location=None, tools="", tooltips="$name: @$name")
        p.vbar_stack(sorted(dataframe[stack_col].unique().astype(str)), x="x", source=pd.DataFrame(data), color=colors, width=0.8, legend_label=legend)
        p.xaxis.axis_label = x_label
        p.yaxis.axis_label = "Counts"
        p.js_on_event(events.DoubleTap, toggle_legend_js(p))
        p.legend.click_policy="hide"
        return p

    if isinstance(data, pd.DataFrame):
        return _stacked_hist(data, x_col, stack_col, x_label, width, height)
    elif isinstance(data, list):
        plot_list = []
        for df in data:
            if link_plots and len(plot_list) > 0:
                plot_list.append(_stacked_hist(df, x_col, stack_col, x_label, width, height, plot_list[0].x_range, plot_list[0].y_range))
            else:
                plot_list.append(_stacked_hist(df, x_col, stack_col, x_label, width, height))
        return plot_list
    else:
        raise TypeError("data has to be of type list or pd.Dataframe")

# Cell
def categorical_2d_histogram(data: pd.DataFrame, category_col: str, hist_col: str, bins=10, range=None, normalize=False, precision=2, color_mapper=None, hist_col_is_categorical=False, width=500, height=500):
    """Creates a 2d histogram with the y axis being a category and the x axis the value each category is histogramed over."""
    def _create_colorbar_and_color_mapper(min_val, max_val, color_mapper):
        # create color bar
        if color_mapper is None:
            color_mapper = LinearColorMapper(palette=Viridis[256], low=min_val, high=max_val)

        color_bar = ColorBar(
            color_mapper=color_mapper, major_label_text_font_size="7px",
            label_standoff=6, border_line_color=None, location=(0, 0)
        )
        return color_bar, color_mapper

    def _categorical_2d_histogram(dataframe, category_col, hist_col, bins, range, normalize, precision, color_mapper, hist_col_is_categorical, width, height):
        data = {"x": np.array([]), "y": [], "z": np.array([])}
        hist_min = dataframe[hist_col].min() if range is None else range[0]
        hist_max = dataframe[hist_col].max() if range is None else range[1]
        if hist_max < hist_min:
            hist_max = hist_min*1.1
        for group in dataframe.groupby(category_col):
            if hist_col_is_categorical:
                bins = dataframe[hist_col].nunique()
                unique_values = dataframe[hist_col].unique()
                x, z = np.unique(group[1][hist_col], return_counts=True)
                data["y"] += [str(group[0])]*bins
                z = np.array([float(z[x == i]) if i in x else 0 for i in unique_values])
                if normalize:
                    z = z/z.sum()
                data["x"] = np.append(data["x"], unique_values.astype(str))
                data["z"] = np.append(data["z"],z)
            else:
                z,x = np.histogram(group[1][hist_col], bins=bins, range=(hist_min, hist_max))
                if normalize:
                    z = z/z.sum()
                data["y"] += [str(group[0])]*bins
                data["x"] = np.append(data["x"],x[:-1].round(precision).astype(str))
                data["z"] = np.append(data["z"],z)
        df = pd.DataFrame(data)

        color_bar, color_mapper = _create_colorbar_and_color_mapper(data["z"].min(), data["z"].max(), color_mapper)

        # create the 2d histogram
        p = figure(
            x_range=sorted(df["x"].unique().astype(str)), y_range=sorted(df["y"].unique().astype(str)),
            x_axis_location="above", tools="hover", toolbar_location=None, tooltips=[('', '@z')],
            width=width, height=height, x_axis_label=hist_col, y_axis_label=category_col
        )
        p.grid.grid_line_color = None
        p.axis.axis_line_color = None
        p.axis.major_tick_line_color = None
        p.axis.major_label_standoff = 0

        p.rect(
            x="x", y="y", width=1, height=1, source=df,
            fill_color={'field': "z", 'transform': color_mapper},
            line_color=None
        )

        p.add_layout(color_bar, 'right')

        return p

    if isinstance(data, pd.DataFrame):
        return _categorical_2d_histogram(data, category_col, hist_col, bins, range, normalize, precision, color_mapper, hist_col_is_categorical, width, height)
    elif isinstance(data, list):
        plot_list = [_categorical_2d_histogram(df, category_col, hist_col, bins, range, normalize, precision, color_mapper, hist_col_is_categorical, width, height) for df in data]
        return plot_list
    else:
        raise TypeError("data has to be of type ")

# Cell
def categorical_2d_histogram_with_gui(data: pd.DataFrame, category_cols=None, hist_cols=None, width=500, height=500):
    """Creates a categorical_2d_histogram for a dataframe, where each option (except width, height and color_mapper) of the categorical_2d_histogram can be set with gui elements.
    If the input is a list all dataframes need to have the same cols as the first dataframe in the list"""
    if category_cols is None and isinstance(data, pd.DataFrame):
        category_cols = data.columns.tolist()
    elif category_cols is None and isinstance(data, list):
        category_cols = data[0].columns.tolist()
    if hist_cols is None and isinstance(data, pd.DataFrame):
        hist_cols = [col_name for col_name in data.columns if is_numeric_dtype(data[col_name])]
    elif hist_cols is None and isinstance(data, list):
        hist_cols = [col_name for col_name in data[0].columns if is_numeric_dtype(data[0][col_name])]
    x_select = pnw.Select(name="X-Axis", options=hist_cols)
    y_select = pnw.Select(name="Y-Axis", options=category_cols)

    axis_selector = pn.Row(x_select, y_select, width=width)

    x_is_categorical = pnw.Checkbox(name="X is categorical", value=False)
    x_bins = pnw.IntInput(name="Bins", start=1, end=500, value=10, disabled=False)
    if isinstance(data, pd.DataFrame):
        x_range_start=data[x_select.value].min()
        x_range_end=data[x_select.value].max()
        x_range_end=x_range_end if x_range_end > x_range_start else x_range_end*1.1
        x_range_step=(x_range_end - x_range_start) / 50
    elif isinstance(data, list):
        x_range_start=min(df[x_select.value].min() for df in data)
        x_range_end=max(df[x_select.value].max() for df in data)
        x_range_end=x_range_end if x_range_end > x_range_start else x_range_end*1.1
        x_range_step=(x_range_end - x_range_start) / 50
    x_range = pnw.RangeSlider(name="X-Axis Range", start=x_range_start, end=x_range_end, step=x_range_step, disabled=False)

    x_axis_configuration = pn.Row(x_range, x_bins, width=width)

    normalize_rows = pnw.Checkbox(name="Normalize rows", value=False)
    x_precision = pnw.IntInput(name="Precision", start=0, end=10, value=2, disabled=False)

    additional_parameters = pn.Row(x_is_categorical, normalize_rows, x_precision, width=width)

    config_gui = pn.Column(axis_selector, additional_parameters, x_axis_configuration, align="center")

    @pn.depends(y_select.param.value, x_select.param.value, x_bins.param.value_throttled, x_range.param.value_throttled, x_is_categorical.param.value, normalize_rows.param.value, x_precision.param.value)
    def _plot(category_col, hist_col, bins, range, x_is_categorical, normalize_rows, precision):
        if x_is_categorical:
            x_range.disabled = True
            x_precision.disabled = True
            x_bins.disabled = True
        else:
            x_range.disabled = False
            x_precision.disabled = False
            x_bins.disabled = False

        if isinstance(data, pd.DataFrame):
            if data[hist_col].min() != x_range.start or data[hist_col].max() != x_range.end:
                x_range.start = data[hist_col].min()
                x_range.end = data[hist_col].max() if data[hist_col].max() > x_range.start else data[hist_col].max()*1.1
                x_range.value = (x_range.start, x_range.end)
                x_range.step = (x_range.end-x_range.start)/50
        elif isinstance(data, list):
            if min(df[hist_col].min() for df in data) != x_range.start or max(df[hist_col].max() for df in data) != x_range.end:
                x_range.start = min(df[hist_col].min() for df in data)
                x_range.end = max(df[hist_col].max() for df in data) if max(df[hist_col].max() for df in data) > x_range.start else max(df[hist_col].max() for df in data)*1.1
                x_range.value = (x_range.start, x_range.end)
                x_range.step = (x_range.end-x_range.start)/50
        range = x_range.value

        if isinstance(data, list):
            plot = pn.Row(*categorical_2d_histogram(data, category_col, hist_col, bins, range, normalize_rows, precision, color_mapper=None, hist_col_is_categorical=x_is_categorical, width=width, height=height))
        elif isinstance(data, pd.DataFrame):
            plot = pn.Row(categorical_2d_histogram(data, category_col, hist_col, bins, range, normalize_rows, precision, color_mapper=None, hist_col_is_categorical=x_is_categorical, width=width, height=height))
        return plot

    return pn.Column(config_gui, _plot)

# Cell
def scatter_plot_with_gui(dataframe: pd.DataFrame, x_cols: list = None, y_cols: list = None, color_cols: list=None, with_hist: bool=True, width=500, height=500):
    x_cols = list(dataframe.columns) if x_cols is None else x_cols
    y_cols = list(dataframe.columns) if y_cols is None else y_cols

    divider = 2

    if color_cols is not None:
        divider = 3
        color_select = pnw.Select(name="Color", options=color_cols, height=50, width=width//divider)

    x_select = pnw.Select(name="X-Axis", options=x_cols, height=50, width=width//divider)
    y_select = pnw.Select(name="Y-Axis", options=y_cols, height=50, width=width//divider)

    if color_cols is None:
        gui = pn.Column(pn.Row(x_select, y_select), None)
    else:
        gui = pn.Column(pn.Row(x_select, y_select, color_select), None)

    def plot(event):
        x_col = x_select.value
        y_col = y_select.value
        df_copy = dataframe.copy()
        if color_cols is not None:
            color_col = color_select.value
            df_copy[color_col] = df_copy[color_col].astype(str)
        source = ColumnDataSource(df_copy)

        if color_cols is not None:
            if 3 <= df_copy[color_col].nunique() <= 20:
                colors = Category20[df_copy[color_col].nunique()]
            else:
                colors = viridis(df_copy[color_col].nunique())
        else:
            colors="navy"
        tooltips = [(color_col, f"@{color_col}")] if color_cols is not None else None

        p = figure(width=width, height=height-50, x_axis_label=x_col, y_axis_label=y_col, tooltips=tooltips)
        # radius = min(df_copy[x_col].max()-df_copy[x_col].min(), df_copy[y_col].max()-df_copy[y_col].min())/75 * max(1, np.log(max(df_copy[x_col].max()-df_copy[x_col].min(), df_copy[y_col].max()-df_copy[y_col].min()))/20)
        if color_cols is not None:
            for index, entry in enumerate(df_copy[color_col].unique()):
                local_source = df_copy[df_copy[color_col] == entry]
                p.scatter(x=x_col, y=y_col, legend_label=entry, source=local_source, color=colors[index])
            p.js_on_event(events.DoubleTap, toggle_legend_js(p))
            p.legend.click_policy="hide"
        else:
            p.scatter(x=x_col, y=y_col, source=source)

        if with_hist:
            hist_x = histogram(df_copy[x_col].values, 30, remove_tools=True, height=height//5, width=width)
            hist_x.x_range = p.x_range
            hist_y = histogram(df_copy[y_col].values, 30, remove_tools=True, orientation="vertical", width=width//5, height=height//7*6)
            hist_y.y_range = p.y_range
            gui[-1] = gridplot([[hist_x, None], [p, hist_y]], merge_tools=True)
        else:
            gui[-1] = p
    plot(None)

    if color_cols is not None:
        color_select.param.watch(plot, "value")
    x_select.param.watch(plot, "value")
    y_select.param.watch(plot, "value")

    return gui

# Cell
def plots_as_matrix(plots, ncols, nrows, width=500, height=500):
    """Takes a list of plots and puts them into a matrix"""
    plot_matrix = pn.GridSpec(ncols=ncols, nrows=nrows, width=width, height=height)
    for index, plot in enumerate(plots):
        row_pos = index//ncols
        col_pos = index%ncols
        if row_pos >= nrows:
            break
        else:
            plot_matrix[row_pos, col_pos] = plot
    return plot_matrix


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/utils.ipynb (unless otherwise specified).

__all__ = ['erles_to_string', 'erles_to_counts_to_utf8', 'string_to_erles', 'correct_mask', 'decorrect_mask']

# Cell
import json
from copy import deepcopy
import numpy as np
from PIL import Image

from icevision.core.mask import EncodedRLEs, MaskArray

from pycocotools import mask as mask_utils

# Cell
def erles_to_string(erles):
    erles_copy = deepcopy(erles)
    erles_copy["counts"] = erles_copy["counts"].decode("utf-8")
    return json.dumps(erles_copy)

# Cell
def erles_to_counts_to_utf8(erles):
    erles_copy = deepcopy(erles)
    for entry in erles_copy:
        entry["counts"] = entry["counts"].decode("utf-8")
    return erles_copy

# Cell
def string_to_erles(erles_string):
    erles = json.loads(erles_string)
    erles["counts"] = erles["counts"].encode()
    return erles

# Cell
def correct_mask(mask_array, pad_x, pad_y, width, height):
    # correct mask
    corrected_mask_array = mask_array.transpose(2, 0, 1)
    if round(pad_x/2) > 0:
        corrected_mask_array=corrected_mask_array[:,:,round(pad_x/2):round(-pad_x/2)]
    if round(pad_y/2) > 0:
        corrected_mask_array=corrected_mask_array[:,round(pad_y/2):round(-pad_y/2),:]
    corrected_mask_array = np.array(Image.fromarray(corrected_mask_array[0,:,:]).resize([width, height], Image.NEAREST))
    corrected_mask_array = np.expand_dims(corrected_mask_array, 0)
    # convert mask array to mask and get erles (only one erles exist!)
    corrected_mask = MaskArray(corrected_mask_array)
    return corrected_mask

# Cell
def decorrect_mask(mask_array, pad_x, pad_y, width, height):
    corrected_mask_array = mask_array.transpose(2, 0, 1)
    # resize
    corrected_mask_array = np.array(Image.fromarray(corrected_mask_array[0,:,:]).resize([width, height], Image.NEAREST))
    corrected_mask_array = np.expand_dims(corrected_mask_array, 0)
    # pad
    corrected_mask_array = np.pad(corrected_mask_array, [[0,0], [pad_y, pad_y], [pad_x, pad_x],])
    corrected_mask = MaskArray(corrected_mask_array)
    return corrected_mask